﻿using System.Collections.Generic;
using System.IO;

namespace Recombobulator
{
    class SR1_StructureArray<T> : SR1_Structure where T : SR1_Structure, new()
    {
        private int[] _dimensions = null;

        private T[] _array = null;

        protected const int _MaxArrayLength = 128;

        public SR1_StructureArray(params int[] dimensions)
        {
            _dimensions = new int[dimensions.Length];
            dimensions.CopyTo(_dimensions, 0);

            int arrayLength = 1;
            foreach (int d in dimensions)
            {
                arrayLength *= d;
            }

            if (arrayLength > 0)
            {
                _array = new T[arrayLength];
                for (int i = 0; i < arrayLength; i++)
                {
                    _array[i] = new T();
                    if (i > _MaxArrayLength)
                    {
                        break;
                    }
                }
            }
        }

        public override List<Property> GetProperties(GoalTracker tracker)
        {
            Properties = new List<Property>();

            if (_array == null)
            {
                return null;
            }

            for (int i = 0; i < _array.Length; i++)
            {
                Property newProp = new Property();
                newProp.TypeName = GetTypeName(false);
                newProp.PropertyName = "";
                newProp.PropertyValue = _array[i].ToString();

                int index = i;
                int d = _dimensions.Length;
                while (d > 0)
                {
                    d--;
                    int subIndex = index % _dimensions[d];
                    index /= _dimensions[d];
                    string indexName = "";
                    indexName += "[";
                    indexName += subIndex.ToString();
                    indexName += "]";
                    indexName += newProp.PropertyName;
                    newProp.PropertyName = indexName;
                }

                //if (!tracker.GoalAchieved)
                //{
                    newProp.Children = _array[i].GetProperties(tracker);
                //}

                Properties.Add(newProp);
                if (i > _MaxArrayLength)
                {
                    break;
                }
            }

            return Properties;
        }

        public override TreeList.Node[] CreatePropertyNodes()
        {
            if (_array == null)
            {
                return null;
            }

            TreeList.Node[] elementNodes = new TreeList.Node[_array.Length];
            for (int i = 0; i < _array.Length; i++)
            {
                string typeName = GetTypeName(false);
                string elementName = "";
                string elementValue = _array[i].ToString();

                int index = i;
                int d = _dimensions.Length;
                while (d > 0)
                {
                    d--;
                    int subIndex = index % _dimensions[d];
                    index /= _dimensions[d];
                    string indexName = "";
                    indexName += "[";
                    indexName += subIndex.ToString();
                    indexName += "]";
                    indexName += elementName;
                    elementName = indexName;
                }

                elementNodes[i] = new TreeList.Node(new string[] { "", typeName, elementName, elementValue });
                elementNodes[i].Nodes.AddRange(_array[i].CreatePropertyNodes());
            }

            return elementNodes;
        }

        public override void Read(SortedDictionary<uint, SR1_Structure> dictionary, BinaryReader reader, bool isMember, GoalTracker tracker)
        {
            Start = (uint)reader.BaseStream.Position;
            IsMember = isMember;

            if (!IsMember)
            {
                dictionary.Add(Start, this);
            }

            if (_array != null)
            {
                for (int i = 0; i < _array.Length; i++)
                {
                    _array[i].Read(dictionary, reader, true, tracker);
                }
            }

            End = (uint)reader.BaseStream.Position;
        }

        public override string ToString()
        {
            if (_array == null)
            {
                return "{ }";
            }

            return "{...}";
        }

        public override string GetTypeName(bool includeDimensions)
        {
            string typeName = typeof(T).Name;

            if (_array == null)
            {
                typeName += "[]";
                return typeName;
            }

            if (includeDimensions)
            {
                int rank = _dimensions.Length;
                for (int r = 0; r < rank; r++)
                {
                    typeName += "[";
                    typeName += _dimensions[r].ToString();
                    typeName += "]";
                }
            }

            return typeName;
        }
    }
}
