#!/usr/bin/env python3

import json
import math
import ntpath
import os
import random
import re
import sys

# Use a fixed random seed for now to ensure consistent results across multiple runs
# The only use of random is to assign model numbers for multi-model objects, which is a 
# temporary hack until I know 
random.seed(26354)

gLineSeparator = "\n"

def showBanner():
	print('Soul Reaver Blender Import/Arrange Generation Script')
	print('Ben Lincoln, 2019-2020')
	print('https://www.thelostworlds.net/')
	
gDigitStrings = ["0", "1", "2", "3", "4", "5", "6", "7", "8", "9"]

class StringElement:
	def __init__(self, originalString):
		self.StringRepresentation = originalString
		self.NumericRepresentation = None
		self.IsNumber = False
		onlyDigitsRegexString = "^[0-9]+$"
		onlyDigitsRegex = re.compile(onlyDigitsRegexString)
		if onlyDigitsRegex.match(originalString):
			self.IsNumber = True
			self.NumericRepresentation = int(originalString)

	def GetSortableValue(self):
		if self.IsNumber:
			return self.NumericRepresentation
		else:
			return self.StringRepresentation

	# returns 0 for =, -1 for less than the other element, +1 for greater than the other element
	def CompareTo(self, otherStringElement, caseSensitive = True):
		selfString = self.StringRepresentation
		otherString = otherStringElement.StringRepresentation
		if not caseSensitive:
			selfString = self.StringRepresentation.lower()
			otherString = otherStringElement.StringRepresentation.lower()
		if self.IsNumber and not otherStringElement.IsNumber:
			return -1
		if otherStringElement.IsNumber and not self.IsNumber:
			return 1
		if self.IsNumber and otherStringElement.IsNumber:
			if self.NumericRepresentation < otherStringElement.NumericRepresentation:
				return -1
			if self.NumericRepresentation > otherStringElement.NumericRepresentation:
				return 1
			# equal numeric values are handled as string comparisons to that e.g. 1, 01, and 001 sort consistently
			if len(self.StringRepresentation) < len(otherStringElement.StringRepresentation):
				return -1
			if len(self.StringRepresentation) > len(otherStringElement.StringRepresentation):
				return 1
		if self.StringRepresentation == otherStringElement.StringRepresentation:
			return 0
		# Use Python's own string-sorting method
		comparisonArray = []
		comparisonArray.append(self.StringRepresentation)
		comparisonArray.append(otherStringElement.StringRepresentation)
		comparisonArray.sort()
		if comparisonArray[0] == self.StringRepresentation:
			return -1
		return 1

class StringWithNumericElements:
	def __init__(self, originalString):
		self.OriginalString = originalString
		self.Elements = []
		doneParsing = False
		if originalString == "":
			self.Elements.append("")
		else:
			currentElementString = ""
			currentPosition = 0
			currentElementIsNumeric = False
			firstChar = True
			lastChar = False
			while not doneParsing:				
				nextChar = None
				nextCharIsDigit = False
				if currentPosition >= len(originalString):
					lastChar = True
					doneParsing = True
					# flip this bit so that the last character in the input string triggers the creation of an element using the current buffer
					nextCharIsDigit = not currentElementIsNumeric
				else:
					nextChar = originalString[currentPosition:currentPosition+1]
					if nextChar in gDigitStrings:
						nextCharIsDigit = True
				if firstChar:
					currentElementString = nextChar
					if not lastChar:
						currentElementIsNumeric = nextCharIsDigit
				if lastChar or not firstChar:
					if currentElementIsNumeric:
						if nextCharIsDigit:
							currentElementString += nextChar
						else:
							e = StringElement(currentElementString)
							self.Elements.append(e)
							currentElementString = nextChar
							currentElementIsNumeric = nextCharIsDigit
					else:
						if nextCharIsDigit:
							e = StringElement(currentElementString)
							self.Elements.append(e)
							currentElementString = nextChar
							currentElementIsNumeric = nextCharIsDigit
						else:
							currentElementString += nextChar
				if firstChar:
					firstChar = False
				currentPosition += 1
				
	def CompareTo(self, otherStringWithNumericElements, caseSensitive = True):
		elementCount = len(self.Elements)
		if len(otherStringWithNumericElements.Elements) < elementCount:
			elementCount = len(otherStringWithNumericElements.Elements)
		for i in range(0, elementCount):
			compValue = self.Elements[i].CompareTo(otherStringWithNumericElements.Elements[i], caseSensitive)
			if compValue == 1:
				return 1
			if compValue == -1:
				return -1
		return 0

class StringWithNumericSuffixCollection:
	def __init__(self):
		self.StringWithNumericElementsArray = []
	
	# basic bubble sort using each element in each string
	def Sort(self, reverse = False, caseSensitive = True):
		if len(self.StringWithNumericElementsArray) < 2:
			return
		doneSorting = False
		while not doneSorting:
			swappedAnyElements = False
			for i in range(0, len(self.StringWithNumericElementsArray) - 1):
				currentElement = self.StringWithNumericElementsArray[i]
				nextElement = self.StringWithNumericElementsArray[i + 1]
				compValue = currentElement.CompareTo(nextElement, caseSensitive)
				if reverse:
					compValue *= -1
				if compValue == 1:
					self.StringWithNumericElementsArray[i] = nextElement
					self.StringWithNumericElementsArray[i + 1] = currentElement
					swappedAnyElements = True
			if not swappedAnyElements:
				doneSorting = True

def SortStringsIntuitively(stringArray):
	coll = StringWithNumericSuffixCollection()
	maxElements = 0
	for s in stringArray:
		newS = StringWithNumericElements(s)
		if len(newS.Elements) > maxElements:
			maxElements = len(newS.Elements)
		coll.StringWithNumericElementsArray.append(newS)
	coll.Sort()
	result = []
	for s in coll.StringWithNumericElementsArray:
		result.append(s.OriginalString)
	return result

class ObjectCounter:
	def __init__(self):
		self.ObjectNumber = 0

class Portal:
	def __init__(self):
		self.Name = ""
		self.NameRaw = ""
		self.SignalNum = -1
		self.DestinationSignalNum = -1
		self.PosX = 0.0
		self.PosY = 0.0
		self.PosZ = 0.0

def GetBlenderName(areaName):
	return "area_%s" % (areaName)

def GetBlenderAreaSetName(areaSetName):
	return "aset_%s" % (areaSetName)
	
def GetBlenderAreaWithObjectsName(areaName):
	return "awo_%s" % (areaName)

class G2EngineObject:
	def __init__(self):
		# Name should be unique, e.g. city2-vlgra-001
		self.Name = ""
		# ObjectTypeName is the name of the object as Soul Reaver thinks of it
		# e.g. vlgra
		self.ObjectTypeName = ""
		self.ObjectNumber = 0
		self.DistanceFromOrigin = 0
		self.ParentAreaName = ""
		self.ParentAreaBlenderName = ""
		self.ColladaFilePath = ""
		self.ObjectNumWithinParent = 0
		self.PosX = 0.0
		self.PosY = 0.0
		self.PosZ = 0.0
		self.OriginalPosX = 0.0
		self.OriginalPosY = 0.0
		self.OriginalPosZ = 0.0
		self.OffsetFromParentAreaX = 0.0
		self.OffsetFromParentAreaY = 0.0
		self.OffsetFromParentAreaZ = 0.0
		self.SpectralPosX = 0.0
		self.SpectralPosY = 0.0
		self.SpectralPosZ = 0.0
		self.OriginalSpectralPosX = 0.0
		self.OriginalSpectralPosY = 0.0
		self.OriginalSpectralPosZ = 0.0
		self.SpectralOffsetFromParentAreaX = 0.0
		self.SpectralOffsetFromParentAreaY = 0.0
		self.SpectralOffsetFromParentAreaZ = 0.0
		self.RotationX = 0.0
		self.RotationY = 0.0
		self.RotationZ = 0.0
		
	def SetProperties(self, objectName, objectColladaDirectoryPath, parentAreaName, objectCounter):
		self.ObjectTypeName = objectName.lower()
		self.ParentAreaName = parentAreaName
		objectCounter.ObjectNumber += 1
		objectNum = objectCounter.ObjectNumber
		self.ColladaFilePath = os.path.join(objectColladaDirectoryPath, "%s.dae" % (self.ObjectTypeName))

	def SetName(self):
		self.Name = "%s_obj_%s_%s" % (self.ParentAreaName, self.ObjectNumWithinParent, self.ObjectTypeName)
		print("%s will be child %s of %s" % (self.Name, self.ObjectNumWithinParent, self.ParentAreaName))
		
	def GetBlenderCode(self, parentObjectPosX, parentObjectPosY, parentObjectPosZ, positionMultiplier, inSpectral, parentAreaWithObjects):
		self.PosX = (float(parentObjectPosX) + float(self.OffsetFromParentAreaX)) * positionMultiplier
		self.PosY = (float(parentObjectPosY) + float(self.OffsetFromParentAreaY)) * positionMultiplier
		self.PosZ = (float(parentObjectPosZ) + float(self.OffsetFromParentAreaZ)) * positionMultiplier
		self.SpectralPosX = (float(parentObjectPosX) + float(self.SpectralOffsetFromParentAreaX)) * positionMultiplier
		self.SpectralPosY = (float(parentObjectPosY) + float(self.SpectralOffsetFromParentAreaY)) * positionMultiplier
		self.SpectralPosZ = (float(parentObjectPosZ) + float(self.SpectralOffsetFromParentAreaZ)) * positionMultiplier
		
		outputRotationX = (float(self.RotationX) / float(4096)) * math.pi * 2.0
		outputRotationY = (float(self.RotationY) / float(4096)) * math.pi * 2.0
		outputRotationZ = (float(self.RotationZ) / float(4096)) * math.pi * 2.0
		
		result = ""
		if self.ColladaFilePath != "" and os.path.isfile(self.ColladaFilePath):
			print("Will set position of %s to %s, %s, %s (spectral %s, %s, %s) and rotation to %s, %s, %s in Blender" % (self.Name, self.PosX, self.PosY, self.PosZ, self.SpectralPosX, self.SpectralPosY, self.SpectralPosZ, outputRotationX, outputRotationY, outputRotationZ))
			
			colladaPath = self.ColladaFilePath.replace("\\", "\\\\")
			posX = self.PosX
			posY = self.PosY
			posZ = self.PosZ
			# I don't think the spectral position is ever used in the data files
			# if inSpectral:
				# posX = self.SpectralPosX
				# posY = self.SpectralPosY
				# posZ = self.SpectralPosZ

			# importObject(objectBaseName, objectColladaPath, parentAreaName, parentAreaWithObjects, parentAreaObjectNumber, placementOrder, distanceFromOrigin, posX, posY, posZ, rotA, rotB, rotC)
			result += 'importObject("%s", "%s", "%s", %s, %s, %s, %s, %s, %s, %s, %s, %s, %s)' % (self.Name, colladaPath, self.ParentAreaName, parentAreaWithObjects, self.ObjectNumWithinParent, self.ObjectNumber, self.DistanceFromOrigin, posX, posY, posZ, outputRotationX, outputRotationY, outputRotationZ)
			result += gLineSeparator
		
		return result

class G2EngineLight:
	def __init__(self):
		self.Name = ""
		self.PosX = 0.0
		self.PosY = 0.0
		self.PosZ = 0.0

class Area:
	def __init__(self):
		self.Name = ""
		self.ObjectNumber = -1
		self.DistanceFromOrigin = 0
		self.AreaSetName = ""
		self.ColladaFilePath = ""
		self.JSONFilePath = ""
		self.PosX = 0.0
		self.PosY = 0.0
		self.PosZ = 0.0
		self.Portals = {}
		self.G2EngineObjects = []
		
	def GetAreaSetName(self, areaName):
		result = re.sub(r'[0-9].*', '', areaName)
		return result
		
	def SetPropertiesFromFile(self, areaName, jsonFilePath, areasColladaDirectoryPath, objectsColladaDirectoryPath, allObjectFilePaths, importObjects, ignoredObjects, objectWhitelist, objectCounter, randomizedObjectNames, placeObjectsSimultaneously = False):
		self.Name = areaName.lower()
		self.AreaSetName = self.GetAreaSetName(self.Name)
		objectCounter.ObjectNumber += 1
		self.ObjectNumber = objectCounter.ObjectNumber
		print("%s will be a child of %s" % (self.Name, self.AreaSetName))
		
		self.JSONFilePath = jsonFilePath
		self.ColladaFilePath = os.path.join(areasColladaDirectoryPath, "%s.dae" % (self.Name))
		inputFile = open(self.JSONFilePath, "r")
		jsonData = json.load(inputFile)
		for topLevelItem in jsonData:
			for level1Prop in topLevelItem['Properties']:
				#print("Type: %s, PropertyName: %s" % (level1Prop['TypeName'], level1Prop['PropertyName']))
				if level1Prop['PropertyName'] == "portals":
					#print("Found portal list")
					
					if 'Children' in level1Prop.keys() and level1Prop['Children'] and len(level1Prop['Children']) > 0:
						for portalEntry in level1Prop['Children']:
							newPortal = Portal()
							if portalEntry['TypeName'] == "StreamUnitPortal":
								if 'Children' in portalEntry.keys() and portalEntry['Children'] and len(portalEntry['Children']) > 0:
									for portalProp in portalEntry['Children']:
										if portalProp['PropertyName'] == "tolevelname":
											newPortal.NameRaw = portalProp['PropertyValue'].lower()
											print("Got portal name %s" % (newPortal.NameRaw))
											npNameSplit = newPortal.NameRaw.split(',')
											newPortal.Name = npNameSplit[0]
											newPortal.DestinationSignalNum = int(npNameSplit[1])
										if portalProp['PropertyName'] == "minx":
											newPortal.PosX = int(portalProp['PropertyValue'])
										if portalProp['PropertyName'] == "miny":
											newPortal.PosY = int(portalProp['PropertyValue'])
										if portalProp['PropertyName'] == "minz":
											newPortal.PosZ = int(portalProp['PropertyValue'])
										if portalProp['PropertyName'] == "MSignalID":
											newPortal.SignalNum = int(portalProp['PropertyValue'])
										
									print("Found portal from %s to %s" % (self.Name, newPortal.NameRaw))
									self.Portals[newPortal.NameRaw] = newPortal
				if importObjects and "Intro" in level1Prop['TypeName']:
					print("Found object entry")
					if 'Children' in level1Prop.keys() and level1Prop['Children'] and len(level1Prop['Children']) > 0:
						print("Found object entry child entries")
						newObjectName = ""
						newObjectTypeName = ""
						newObjectParentAreaName = ""
						newObjectDistanceFromOrigin = 0
						newObjectParentAreaBlenderName = ""
						newObjectObjectNumWithinParent = 0
						newObjectRotationX = 0.0
						newObjectRotationY = 0.0
						newObjectRotationZ = 0.0
						newObjectOffsetFromParentAreaX = 0.0
						newObjectOffsetFromParentAreaY = 0.0
						newObjectOffsetFromParentAreaZ = 0.0
						newObjectSpectralOffsetFromParentAreaX = 0.0
						newObjectSpectralOffsetFromParentAreaY = 0.0
						newObjectSpectralOffsetFromParentAreaZ = 0.0
						
						newObjectColladaPaths = []
						
						continueAddingObject = True
						for objectEntry in level1Prop['Children']:
							if continueAddingObject:
								if objectEntry['PropertyName'] == "name":
									newObjectTypeName = objectEntry['PropertyValue'].lower()
									#print("Found object name '%s'" % (newObjectTypeName))
									candidateFiles = []
									objectSingleModelName = "%s.dae" % (newObjectTypeName)
									objectMultiModelRegex = "%s-model_[0-9]{4}.dae" % (newObjectTypeName)
									for objectFilePath in allObjectFilePaths:
										objectFileName = ntpath.basename(objectFilePath)
										if objectFileName == objectSingleModelName:
											candidateFiles = []
											candidateFiles.append(objectFilePath)
											break
										if re.search(objectMultiModelRegex, objectFileName):
											candidateFiles.append(objectFilePath)
									if len(candidateFiles) == 0:
										print("Error: could not find any Collada files for object '%s'" % newObjectTypeName)
										continueAddingObject = False
										break
									else:
										if len(candidateFiles) == 1:
											newObjectColladaPaths.append(candidateFiles[0])
										else:
											# I don't know what controls the model that's used yet, so just pseudo-randomize them for now
											# As long as the object type is one where only one model should be used
											if newObjectTypeName in randomizedObjectNames:
												candidateFiles.sort()
												newObjectColladaPaths.append(candidateFiles[random.randint(0, len(candidateFiles) - 1)])
											else:
												newObjectColladaPaths = candidateFiles
											
									if continueAddingObject:
										newObjectParentAreaName = self.Name
										newObjectDistanceFromOrigin = self.DistanceFromOrigin + 1
										newObjectParentAreaBlenderName = GetBlenderName(self.Name)
										newObjectObjectNumWithinParent = len(self.G2EngineObjects) + 1
										newObjectName = "%s_obj_%s_%s" % (newObjectParentAreaName, newObjectObjectNumWithinParent, newObjectTypeName)
								if objectEntry['PropertyName'] == "rotation":
									if 'Children' in objectEntry.keys() and objectEntry['Children'] and len(objectEntry['Children']) > 0:
										for rotationEntry in objectEntry['Children']:
											if rotationEntry['PropertyName'] == "x":
												newObjectRotationX = int(rotationEntry['PropertyValue'])
											if rotationEntry['PropertyName'] == "y":
												newObjectRotationY = int(rotationEntry['PropertyValue'])
											if rotationEntry['PropertyName'] == "z":
												newObjectRotationZ = int(rotationEntry['PropertyValue'])
								if objectEntry['PropertyName'] == "position":
									if 'Children' in objectEntry.keys() and objectEntry['Children'] and len(objectEntry['Children']) > 0:
										for positionEntry in objectEntry['Children']:
											if positionEntry['PropertyName'] == "x":
												newObjectOffsetFromParentAreaX = int(positionEntry['PropertyValue'])
											if positionEntry['PropertyName'] == "y":
												newObjectOffsetFromParentAreaY = int(positionEntry['PropertyValue'])
											if positionEntry['PropertyName'] == "z":
												newObjectOffsetFromParentAreaZ = int(positionEntry['PropertyValue'])
								if objectEntry['PropertyName'] == "spectralPosition":
									if 'Children' in objectEntry.keys() and objectEntry['Children'] and len(objectEntry['Children']) > 0:
										for positionEntry in objectEntry['Children']:
											if positionEntry['PropertyName'] == "x":
												newObjectSpectralOffsetFromParentAreaX = int(positionEntry['PropertyValue'])
											if positionEntry['PropertyName'] == "y":
												newObjectSpectralOffsetFromParentAreaY = int(positionEntry['PropertyValue'])
											if positionEntry['PropertyName'] == "z":
												newObjectSpectralOffsetFromParentAreaZ = int(positionEntry['PropertyValue'])
						if continueAddingObject:
							objectDescription = "%s (instance of %s) with position offset %s, %s, %s, spectral position offset %s, %s, %s, and rotation %s, %s, %s" % (newObjectName, newObjectTypeName, newObjectOffsetFromParentAreaX, newObjectOffsetFromParentAreaY, newObjectOffsetFromParentAreaZ, newObjectSpectralOffsetFromParentAreaX, newObjectSpectralOffsetFromParentAreaY, newObjectSpectralOffsetFromParentAreaZ, newObjectRotationX, newObjectRotationY, newObjectRotationZ)
							if not placeObjectsSimultaneously:
								objectCounter.ObjectNumber += 1
							#if len(objectWhitelist) > 0:
							#	if newObject.ObjectTypeName not in objectWhitelist:
							#		print("Did not add %s to the objects list of %s, because its type is not in the whitelist" % (objectDescription, self.Name))
							#		continueAddingObject = False
							if newObjectTypeName in ignoredObjects:
								continueAddingObject = False
								print("Did not add %s to the objects list of %s, because its type is in the list to ignore" % (objectDescription, self.Name))
							if continueAddingObject:
								for ocpn in range(0, len(newObjectColladaPaths)):
									newObject = G2EngineObject()
									newObject.ObjectNumber = objectCounter.ObjectNumber
									newObject.ObjectTypeName = newObjectTypeName
									newObject.ParentAreaName = newObjectParentAreaName
									newObject.DistanceFromOrigin = newObjectDistanceFromOrigin
									newObject.ParentAreaBlenderName = newObjectParentAreaBlenderName
									newObject.ObjectNumWithinParent = newObjectObjectNumWithinParent
									newObject.Name = newObjectName
									newObject.RotationX = newObjectRotationX
									newObject.RotationY = newObjectRotationY
									newObject.RotationZ = newObjectRotationZ
									newObject.OffsetFromParentAreaX = newObjectOffsetFromParentAreaX
									newObject.OffsetFromParentAreaY = newObjectOffsetFromParentAreaY
									newObject.OffsetFromParentAreaZ = newObjectOffsetFromParentAreaZ
									newObject.SpectralOffsetFromParentAreaX = newObjectSpectralOffsetFromParentAreaX
									newObject.SpectralOffsetFromParentAreaY = newObjectSpectralOffsetFromParentAreaY
									newObject.SpectralOffsetFromParentAreaZ = newObjectSpectralOffsetFromParentAreaZ
									newObject.ColladaFilePath = newObjectColladaPaths[ocpn]
									self.G2EngineObjects.append(newObject)
									print("Added %s (path '%s') to the objects list of %s" % (objectDescription, newObject.ColladaFilePath, self.Name))
							#if 'Children' in objectEntry.keys() and objectEntry['Children'] and len(objectEntry['Children']) > 0:
					
		
	def GetBlenderCode(self, positionMultiplier, inSpectral):
		outputX = float(self.PosX) * positionMultiplier
		outputY = float(self.PosY) * positionMultiplier
		outputZ = float(self.PosZ) * positionMultiplier
		print("Will set position of %s to %s, %s, %s in Blender" % (self.Name, outputX, outputY, outputZ))

		colladaPath = self.ColladaFilePath.replace("\\", "\\\\")

		result = ""
		#importArea(areaSetBaseName, areaName, areaColladaPath, placementOrder, distanceFromOrigin, posX, posY, posZ)
		result += 'awo = importArea("%s", "%s", "%s", %s, %s, %s, %s, %s, debugLogPath = "%s")' %(self.AreaSetName, self.Name, colladaPath, self.ObjectNumber, self.DistanceFromOrigin, outputX, outputY, outputZ, logPath.replace("\\", "\\\\"))
		result += gLineSeparator
		
		for o in self.G2EngineObjects:
			if o.Name != "":
				o.DistanceFromOrigin = self.DistanceFromOrigin + 1
				result += o.GetBlenderCode(self.PosX, self.PosY, self.PosZ, positionMultiplier, inSpectral, "awo")
		
		return result


def WriteToFile(outputFilePath, content):
	outputFile = open(outputFilePath, "a", newline=gLineSeparator)
	outputFile.write(content)
	outputFile.write(gLineSeparator)
	outputFile.close()


def AreaFromFile(areaName, inputFilePath, areasColladaDirectoryPath, objectsColladaDirectoryPath, allObjectFilePaths, importObjects, ignoredObjects, objectWhitelist, objectCounter, randomizedObjectNames, placeObjectsSimultaneously = False):
	newArea = Area()
	newArea.SetPropertiesFromFile(areaName, inputFilePath, areasColladaDirectoryPath, objectsColladaDirectoryPath, allObjectFilePaths, importObjects, ignoredObjects, objectWhitelist, objectCounter, randomizedObjectNames, placeObjectsSimultaneously = placeObjectsSimultaneously)
	return newArea
							

class ProcessingTracker:
	def __init__(self):
		self.Counter = ObjectCounter()
		self.MaxDistanceFromOrigin = 0
		self.Areas = {}
		self.AreaNames = []
		self.AreaSetNames = []
		self.G2EngineObjects = {}
		self.ObjectNames = []
		self.BreadthFirst = True
		self.ImportObjects = False
		self.ImportLights = False
		self.AreaJSONDirectory = ""
		self.AreaColladaDirectory = ""
		self.ObjectColladaDirectory = ""
		# for tracking unlinked areas and objects
		self.AllAreaColladaFiles = []
		self.UsedAreaColladaFiles = []
		self.AllObjectColladaFiles = []
		self.UsedObjectColladaFiles = []
		self.AreasToProcess = []
		self.IgnoredObjectNames = []
		self.ObjectNameWhitelist = []
		self.AreaWhitelistRegex = None
		self.PortalOverrides = {}
		self.BlacklistedPortalCombinations = []
		self.SpectralMode = False
		self.SpectralLightingMode = False
		self.MaterialMode = False
		self.ConfirmPositionOfMultiLinkedAreas = False
		self.PlaceObjectsSimultaneously = False
		self.PotentiallyOrphanedAreas = {}
		self.BossObjectNames = [ "alukabss", "kain", "morboss", "roninbss", "skinbos", "walboss", "walbosa", "walbosb", "walbosc", "walbosd", "walbose" ]
		self.MaterialPlaneMonsterObjectNames = [ "aluka", "human", "hunter", "morlock", "priests", "raven", "ronin", "skinner", "vlgra", "vlgrb", "vlgrc", "wallcr", "wrshp", "xbow" ]
		self.MaterialPlaneOtherObjectNames = [ "healthr" ]
		self.SpectralPlaneMonsterObjectNames = [ "sluagh", "soul", "vwraith" ]
		self.SpectralPlaneOtherObjectNames = [ "portal" ]
		self.NonVisibleObjectNames = [ "boshost", "campath", "cathost", "comblst", "comcath", "comet", "comleft", "comlftb", "comritb", "comrite", "cronhstk", "cronhstr", "dummy", "dust", "eaggot", "eaggotr", "eaggots", "eaggotu", "eggfx", "firhost", "flamea", "flameb", "flamec", "flamed", "flamegs",  "glfhost", "hint", "host", "hostfor", "marker", "pilhost", "pilhostk", "pilxhst", "ronhost", "saggot", "saggots", "sfxclwf", "sfxfalb", "sfxfall", "sfxgear", "sfxpipe", "sfxrush", "sfxvort", "sfxwinb", "sfxwinc", "sfxwood", "sfxworm", "sknhost", "spldum", "spldus", "splob", "tmbhost", "tranhst", "undhost", "updraft", "walhostb", "walsoul", "warpg", "wportal" ]
		self.RandomizedObjectNames = [ "aluka", "banner", "flagall", "hunter", "morlock", "ronin", "skinner", "stnblk", "vlgr", "vlgra", "vlgrb", "vlgrc", "wrshp", "xbow" ]
	
	def AppendList(self, existingList, newList):
		result = []
		for eli in existingList:
			result.append(eli)
		for nli in newList:
			result.append(nli)
		return result
	
	def GetAreaToAreaString(self, area1, area2):
		result = "%s_to_%s" % (area1, area2)
		return result
		
	def RecursePortals(self, currentArea, distanceFromOrigin):
		if currentArea.Name in self.AreaNames:
			print("Ignoring area %s, as it has already been processed" % (currentArea.Name))
		else:
			if self.AreaWhitelistRegex is not None and not re.match(self.AreaWhitelistRegex, currentArea.Name):
				print("Ignoring %s because it does not match the whitelist regex" % (currentArea.Name))
			else:
				print("Processing %s" % (currentArea.Name))
				if distanceFromOrigin > self.MaxDistanceFromOrigin:
					self.MaxDistanceFromOrigin = distanceFromOrigin
				self.AreaNames.append(currentArea.Name)
				self.Areas[currentArea.Name] = currentArea
				if currentArea.AreaSetName not in self.AreaSetNames:
					self.AreaSetNames.append(currentArea.AreaSetName)
				if currentArea.ColladaFilePath not in self.UsedAreaColladaFiles:
					cfp = currentArea.ColladaFilePath.replace("\\\\", "\\")
					self.UsedAreaColladaFiles.append(cfp)

				portalNamesRaw = []
				portalNames = []
				usedForeignPortals = []
				
				for portalNameRaw in currentArea.Portals.keys():
					#print(portalNameRaw)
					portal = currentArea.Portals[portalNameRaw]
					if portal.Name in self.AreaNames:
						print("Ignoring portal from %s to %s, as %s has already been processed" % (currentArea.Name, portal.Name, portal.Name))
					else:
						portalNamesRaw.append(portal.NameRaw)
						if portal.Name in portalNames:
							print("Warning: area '%s' contains multiple portals to area '%s'. You may need to manually correct the output data for these two areas." % (currentArea.Name, portal.Name))
						else:
							portalNames.append(portal.Name)
						linkedAreaFilePath = os.path.join(self.AreaJSONDirectory, "%s.json" % (portal.Name))
						if not os.path.isfile(linkedAreaFilePath):
							print("Error: could not find the file '%s' expected due to a portal in area '%s'" % (linkedAreaFilePath, currentArea.Name))
						else:
							print("Loading data for area %s from '%s'" % (portal.Name, linkedAreaFilePath))
							linkedArea = AreaFromFile(portal.Name, linkedAreaFilePath, self.AreaColladaDirectory, self.ObjectColladaDirectory, self.AllObjectColladaFiles, self.ImportObjects, self.IgnoredObjectNames, self.ObjectNameWhitelist, self.Counter, self.RandomizedObjectNames, placeObjectsSimultaneously = self.PlaceObjectsSimultaneously)
							foundMatchingPortal = False
							matchingPortals = []
							exactPortalMatches = []
							destinationPortal = None
							for foreignPortalName in linkedArea.Portals.keys():								
								foreignPortal = linkedArea.Portals[foreignPortalName]
								# if foreignPortal.NameRaw in usedForeignPortals:
									# print("Ignoring duplicate reference to %s" % (foreignPortal.NameRaw))
								# else:
									# if foreignPortal.Name == currentArea.Name:
										# matchingPortals.append(foreignPortal)
								if foreignPortal.Name == currentArea.Name:
									blCheckString1 = "%s:%s" % (portal.NameRaw, foreignPortal.NameRaw)
									blCheckString2 = "%s:%s" % (foreignPortal.NameRaw, portal.NameRaw)
									considerMatch = True
									if blCheckString1 in self.BlacklistedPortalCombinations:
										considerMatch = False
									if considerMatch:
										if blCheckString2 in self.BlacklistedPortalCombinations:
											considerMatch = False
									if considerMatch:
										matchingPortals.append(foreignPortal)
										if foreignPortal.SignalNum == portal.DestinationSignalNum:
											print("Found exact portal match: %s to %s" % (portal.NameRaw, foreignPortal.NameRaw))
											exactPortalMatches.append(foreignPortal)
									else:
										print("Ignoring the blacklisted portal combination '%s'" % (blCheckString1))
							if len(exactPortalMatches) > 0:
								foundMatchingPortal = True
								destinationPortal = exactPortalMatches[0]
							if not foundMatchingPortal:
								if len(matchingPortals) > 0:
									print("Warning: no exact matching portal links found from %s to %s" % (currentArea.Name, linkedArea.Name))
									destinationPortal = matchingPortals[0]
									if len(matchingPortals) > 1:
										print("Warning: Multiple inexact portal matches link %s to %s" % (currentArea.Name, linkedArea.Name))
										a2a = self.GetAreaToAreaString(currentArea.Name, linkedArea.Name)
										if a2a in self.PortalOverrides:
											print("\tUsing manual override for %s" % (a2a))
											destinationPortal = matchingPortals[self.PortalOverrides[a2a]]
										else:
											print("\tNo manual override found for %s, will use first common portal" % (a2a))
									foundMatchingPortal = True
							if foundMatchingPortal:
								print("Found other end of portal between %s and %s" % (currentArea.Name, linkedArea.Name))
								foundMatchingPortal = True
								usedForeignPortals.append(destinationPortal.NameRaw)
								diffX = portal.PosX - destinationPortal.PosX
								diffY = portal.PosY - destinationPortal.PosY
								diffZ = portal.PosZ - destinationPortal.PosZ
								linkedArea.PosX = currentArea.PosX + diffX
								linkedArea.PosY = currentArea.PosY + diffY
								linkedArea.PosZ = currentArea.PosZ + diffZ
								print("Offset of %s from %s is %s, %s, %s" % (linkedArea.Name, currentArea.Name, diffX, diffY, diffZ))
				
							if not foundMatchingPortal:
								print("Warning: did not find the other end of portal between %s and %s - manual repositioning may be required. Adding %s to the list of potentially-orphaned objects." % (currentArea.Name, linkedArea.Name, linkedArea.Name))
								print("Using position of %s for %s, as the other end of the portal was not found" % (currentArea.Name, linkedArea.Name))
								self.PotentiallyOrphanedAreas[linkedArea.Name] = linkedArea
								linkedArea.PosX = currentArea.PosX
								linkedArea.PosY = currentArea.PosY
								linkedArea.PosZ = currentArea.PosZ
							
							if foundMatchingPortal:
								if linkedArea.Name in self.PotentiallyOrphanedAreas.keys():
									print("Found better portal match for potentially-orphaned area %s - removing it from the list of potential orphans" % (linkedArea.Name))
									del self.PotentiallyOrphanedAreas[linkedArea.Name]
								linkedArea.DistanceFromOrigin = distanceFromOrigin + 1
								if self.BreadthFirst:
									self.AreasToProcess.append(linkedArea)
								else:
									self.RecursePortals(linkedArea, distanceFromOrigin + 1)

def ShowUsage():
	#print('Usage: %s --start-area-name FIRST_AREA --start-area-file FIRST_AREA_JSON_FILE --areas-collada-dir AREAS_COLLADA_DIRECTORY [--objects-collada-dir OBJECTS_COLLADA_DIRECTORY] --output OUTPUT_BLENDER_SCRIPT [--start-x FIRST_AREA_X_POSITION] [--start-y FIRST_AREA_Y_POSITION] [--start-z FIRST_AREA_Z_POSITION] [--breadth-first|--depth-first] [--import-objects] [--import-lights] [--spectral] [--ignore-raziel] [--ignore-bosses] [--ignore-monsters] [--ignore-non-visible] [--first-object-number NUMBER] [--initial-coordinates X,Y,Z]' % (sys.argv[0]))
	print('Usage: %s --start-area-name FIRST_AREA --start-area-file FIRST_AREA_JSON_FILE --areas-collada-dir AREAS_COLLADA_DIRECTORY [--objects-collada-dir OBJECTS_COLLADA_DIRECTORY] --output OUTPUT_BLENDER_SCRIPT [--breadth-first|--depth-first] [--import-objects] [--import-lights] [--spectral] [--ignore-raziel] [--ignore-bosses] [--ignore-monsters] [--ignore-non-visible] [--initial-object-number NUMBER] [--initial-coordinates X,Y,Z]' % (sys.argv[0]))
	print(gLineSeparator)
	('Example: %s --start-area-name under1 --start-area-file C:\\Soul_Reaver\\JSON\\under1.json --areas-collada-dir C:\\Soul_Reaver\\Areas --objects-collada-dir C:\\Soul_Reaver\\Objects --output soul_reaver_import.py --breadth-first --import-objects --ignore-raziel' % (sys.argv[0]))
	print(gLineSeparator)


def HandleNewArea(pTracker):
	doneLoopProcessing = False
	while not doneLoopProcessing:
		newAreaCount = 0
		# handle breadth-first mode
		doneProcessing = False
		if len(pTracker.AreasToProcess) == 0:
			doneProcessing = True
		else:
			while not doneProcessing:
				nextArea = pTracker.AreasToProcess.pop(0)
				pTracker.RecursePortals(nextArea, nextArea.DistanceFromOrigin + 1)			
				if len(pTracker.AreasToProcess) == 0:
					doneProcessing = True
		
		# handle first stage of any new orphaned areas
		doneProcessing = False
		if len(pTracker.PotentiallyOrphanedAreas.keys()) == 0:
			doneProcessing = True
		else:
			oaNames = []
			for n in pTracker.PotentiallyOrphanedAreas.keys():
				oaNames.append(n)
			while not doneProcessing:
				newAreaCount += 1
				orphanedAreaName = oaNames.pop(0)
				print("Processing orphaned area %s" % (orphanedAreaName))
				nextArea = pTracker.PotentiallyOrphanedAreas[orphanedAreaName]
				pTracker.RecursePortals(nextArea, nextArea.DistanceFromOrigin + 1)	
				del pTracker.PotentiallyOrphanedAreas[orphanedAreaName]
				if len(oaNames) == 0:
					doneProcessing = True
		if newAreaCount == 0:
			doneLoopProcessing = True
			print("Done checking for additional areas")
		else:
			print("Found %s new area(s) - checking for more" % (newAreaCount))
	
if __name__=='__main__':
	positionMultiplier = 0.001
	frameStep = 1
	frameNumber = 1
	colladaExtension = ".dae"
	startArea = ""
	inputFilePath = ""
	outputFilePath = ""
	areasColladaDirectoryPath = ""
	objectsColladaDirectoryPath = ""
	startAreaX = None
	startAreaY = None
	startAreaZ = None
	pTracker = ProcessingTracker()
	pTracker.BreadthFirst = True
	portalOverrideList = None
	initialDistanceFromOrigin = 0
	vertexColourBlend = 0.85
	placeAreas = True
	placeUnlinkedAreas = False
	placeUnusedObjects = False
#	unlinkedAreaMinX = -65000
#	unlinkedAreaMaxX = 187000
	unlinkedAreaMinX = -100000
	unlinkedAreaMaxX = 100000
	unlinkedAreaStepX = 40000
	unlinkedAreaStartY = -150000
	unlinkedAreaStepY = -40000
	unlinkedAreaStartZ = 0
	unlinkedAreaStepZ = 0
	unusedObjectStepX = 8000
	unusedObjectStepY = -8000
	unusedObjectStepZ = 0
	logPath = ""
	addOverheadLights = False
	overheadLightMinX = -160
	overheadLightMaxX = 160
	overheadLightMinY = -160
	overheadLightMaxY = 160
	overheadLightStepX = 80
	overheadLightStepY = 80
	overheadLightZ = 100
	overheadLightStrength = 0.5
	
	addSpinCam = False
	spinCamNumRotations = 3
	spinCamStartHeight = 0
	spinCamEndHeight = 100
	spinCamStartDistance = 1
	spinCamEndDistance = 300
	spinCamRotationFrameCount = 1200
	exponentialSpin = False
	
	addTripleTierSpinCam = False
	ttSpinCamTier1Distance = 100.0
	ttSpinCamTier2Distance = 200.0
	ttSpinCamTier3Distance = 100.0
	ttSpinCamTier1Height = 100.0
	ttSpinCamTier2Height = 0.0
	ttSpinCamTier3Height = -100.0
	
	modelExTestCaseMode = False
	emulateBackfaceCulling = True
	everythingGlows = False
	renderEngine = 'CYCLES'

	# don't process warpgate portals
	pTracker.AreaNames.append('warpgate')
	
	showBanner()
	if len(sys.argv) < 9:
		ShowUsage()
		sys.exit(1)
	
	#baseScriptPath = os.path.dirname(os.path.abspath(inspect.getfile(inspect.currentframe())))
	
	args2 = []
	
	argNum = 0
	while argNum < len(sys.argv):
		currentArg = sys.argv[argNum]
		foundArg = False
		if argNum < (len(sys.argv) - 1):
			nextArg = sys.argv[argNum + 1]
			if currentArg == "--start-area-file":
				foundArg = True
				inputFilePath = os.path.abspath(nextArg)
			if foundArg == False:
				if currentArg == "--start-area-name":
					foundArg = True
					startArea = nextArg
			if foundArg == False:
				if currentArg == "--areas-collada-dir":
					foundArg = True
					areasColladaDirectoryPath = os.path.abspath(nextArg)	
			if foundArg == False:
				if currentArg == "--objects-collada-dir":
					foundArg = True
					objectsColladaDirectoryPath = os.path.abspath(nextArg)			
			if foundArg == False:
				if currentArg == "--output":
					foundArg = True
					outputFilePath = os.path.abspath(nextArg)
			if foundArg == False:
				if currentArg == "--log":
					foundArg = True
					logPath = os.path.abspath(nextArg)
			# if foundArg == False:
				# if currentArg == "--start-x":
					# foundArg = True
					# startAreaX = int(nextArg)
			# if foundArg == False:
				# if currentArg == "--start-y":
					# foundArg = True
					# startAreaY = int(nextArg)
			# if foundArg == False:
				# if currentArg == "--start-z":
					# foundArg = True
					# startAreaZ = int(nextArg)
			if foundArg == False:
				if currentArg == "--area-whitelist":
					foundArg = True
					pTracker.AreaWhitelistRegex = nextArg
			if foundArg == False:
				if currentArg == "--portal-override-list":
					foundArg = True
					portalOverrideList = nextArg
			if foundArg == False:
				if currentArg == "--initial-object-number":
					foundArg = True
					try:
						pTracker.Counter.ObjectNumber = int(nextArg) - 1
					except Exception as e:
						print("Could not parse '%s' as an integer" % (nextArg))
						sys.exit(1)
			if foundArg == False:
				if currentArg == "--initial-distance-from-origin":
					foundArg = True
					try:
						initialDistanceFromOrigin = int(nextArg) - 1
					except Exception as e:
						print("Could not parse '%s' as an integer" % (nextArg))
						sys.exit(1)
			if foundArg == False:
				if currentArg == "--vertex-blend":
					foundArg = True
					try:
						vertexColourBlend = float(nextArg)
					except Exception as e:
						print("Could not parse '%s' as a floating-point number" % (nextArg))
						sys.exit(1)
			if foundArg == False:
				if currentArg == "--spin-num-rotations":
					foundArg = True
					try:
						spinCamNumRotations = float(nextArg)
					except Exception as e:
						print("Could not parse '%s' as a floating-point number" % (nextArg))
						sys.exit(1)
			if foundArg == False:
				if currentArg == "--spin-start-height":
					foundArg = True
					try:
						spinCamStartHeight = float(nextArg)
					except Exception as e:
						print("Could not parse '%s' as a floating-point number" % (nextArg))
						sys.exit(1)
			if foundArg == False:
				if currentArg == "--spin-end-height":
					foundArg = True
					try:
						spinCamEndHeight = float(nextArg)
					except Exception as e:
						print("Could not parse '%s' as a floating-point number" % (nextArg))
						sys.exit(1)
			if foundArg == False:
				if currentArg == "--spin-start-distance":
					foundArg = True
					try:
						spinCamStartDistance = float(nextArg)
					except Exception as e:
						print("Could not parse '%s' as a floating-point number" % (nextArg))
						sys.exit(1)
			if foundArg == False:
				if currentArg == "--spin-end-distance":
					foundArg = True
					try:
						spinCamEndDistance = float(nextArg)
					except Exception as e:
						print("Could not parse '%s' as a floating-point number" % (nextArg))
						sys.exit(1)
			if foundArg == False:
				if currentArg == "--spin-num-frames":
					foundArg = True
					try:
						spinCamRotationFrameCount = int(nextArg)
					except Exception as e:
						print("Could not parse '%s' as an integer" % (nextArg))
						sys.exit(1)
			if foundArg == False:
				if currentArg == "--tt-spin-distance1":
					foundArg = True
					try:
						ttSpinCamTier1Distance = float(nextArg)
					except Exception as e:
						print("Could not parse '%s' as a floating-point number" % (nextArg))
						sys.exit(1)
			if foundArg == False:
				if currentArg == "--tt-spin-distance2":
					foundArg = True
					try:
						ttSpinCamTier2Distance = float(nextArg)
					except Exception as e:
						print("Could not parse '%s' as a floating-point number" % (nextArg))
						sys.exit(1)
			if foundArg == False:
				if currentArg == "--tt-spin-distance3":
					foundArg = True
					try:
						ttSpinCamTier3Distance = float(nextArg)
					except Exception as e:
						print("Could not parse '%s' as a floating-point number" % (nextArg))
						sys.exit(1)
			if foundArg == False:
				if currentArg == "--tt-spin-height1":
					foundArg = True
					try:
						ttSpinCamTier1Height = float(nextArg)
					except Exception as e:
						print("Could not parse '%s' as a floating-point number" % (nextArg))
						sys.exit(1)
			if foundArg == False:
				if currentArg == "--tt-spin-height2":
					foundArg = True
					try:
						ttSpinCamTier2Height = float(nextArg)
					except Exception as e:
						print("Could not parse '%s' as a floating-point number" % (nextArg))
						sys.exit(1)
			if foundArg == False:
				if currentArg == "--tt-spin-height3":
					foundArg = True
					try:
						ttSpinCamTier3Height = float(nextArg)
					except Exception as e:
						print("Could not parse '%s' as a floating-point number" % (nextArg))
						sys.exit(1)
			if foundArg == False:
				if currentArg == "--initial-coordinates":
					foundArg = True
					coords = nextArg.split(',')
					if len(coords) != 3:
						print("The value for --initial-coordinates should be three floating-point values in the form of X,Y,Z, e.g 100.0,50.0,10.0")
						sys.exit(1)
					try:
						startAreaX = float(coords[0].strip())
					except Exception as e:
						print("Could not parse '%s' as a floating-point number" % (nextArg))
						sys.exit(1)
					try:
						startAreaY = float(coords[1].strip())
					except Exception as e:
						print("Could not parse '%s' as a floating-point number" % (nextArg))
						sys.exit(1)
					try:
						startAreaZ = float(coords[2].strip())
					except Exception as e:
						print("Could not parse '%s' as a floating-point number" % (nextArg))
						sys.exit(1)
			if foundArg:
				argNum += 2
		if foundArg == False:
			args2.append(currentArg)
			argNum += 1

	for a in args2:
		if a == "--no-areas":
			placeAreas = False
		if a == "--ignore-raziel":
			pTracker.IgnoredObjectNames.append("raziel")
		if a == "--ignore-bosses":
			pTracker.IgnoredObjectNames = pTracker.AppendList(pTracker.IgnoredObjectNames, pTracker.BossObjectNames)
		if a == "--ignore-monsters":
			pTracker.IgnoredObjectNames = pTracker.AppendList(pTracker.IgnoredObjectNames, pTracker.MaterialPlaneMonsterObjectNames)
			pTracker.IgnoredObjectNames = pTracker.AppendList(pTracker.IgnoredObjectNames, pTracker.SpectralPlaneMonsterObjectNames)
		if a == "--ignore-non-visible":
			pTracker.IgnoredObjectNames = pTracker.AppendList(pTracker.IgnoredObjectNames, pTracker.NonVisibleObjectNames)
		if a == "--breadth-first":
			pTracker.BreadthFirst = True
		if a == "--depth-first":
			pTracker.BreadthFirst = False
		if a == "--import-objects":
			pTracker.ImportObjects = True
		if a == "--import-lights":
			pTracker.ImportLights = True
		if a == "--spectral":
			pTracker.SpectralMode = True
		if a == "--material":
			pTracker.MaterialMode = True
		# this is separate from the rest of the Spectral mode because it makes the world hard to see when really zoomed out
		if a == "--spectral-lighting":
			pTracker.SpectralLightingMode = True
		if a == "--place-unlinked-areas":
			placeUnlinkedAreas = True
		if a == "--place-unused-objects":
			placeUnusedObjects = True
		if a == "--place-objects-simultaneously":
			pTracker.PlaceObjectsSimultaneously = True
		if a == "--confirm-area-positions":
			pTracker.ConfirmPositionOfMultiLinkedAreas = True
		if a == "--markers-and-campaths":
			pTracker.ObjectNameWhitelist.append("marker")
			pTracker.ObjectNameWhitelist.append("campath")
		if a == "--add-spincam":
			addSpinCam = True
		if a == "--add-tt-spincam":
			addTripleTierSpinCam = True
		if a == "--exponential-spin":
			exponentialSpin = True
		if a == "--add-overhead-lights":
			addOverheadLights = True
		if a == "--overhead-lights-half-strength":
			overheadLightStepX = overheadLightStepX * 2
			overheadLightStepY = overheadLightStepY * 2
		if a == "--modelex-test-cases":
			modelExTestCaseMode = True
		if a == "--double-sided-polygons":	
			emulateBackfaceCulling = False
		if a == "--everything-glows":	
			everythingGlows = True
		if a == "--cycles":	
			renderEngine = 'CYCLES'
		if a == "--eevee":	
			renderEngine = 'BLENDER_EEVEE'
		if a == "--help" or a == "-h" or a == "-?":
			ShowUsage()
			sys.exit(0)

	showUsage = False
	
	if outputFilePath == "":
		showUsage = True

	if placeAreas:
		if inputFilePath == "" or startArea == "" or areasColladaDirectoryPath == "":
			showUsage = True
			
	if pTracker.ImportObjects:
		if objectsColladaDirectoryPath == "":
			showUsage = True			

	if addTripleTierSpinCam and addSpinCam:
		print("Only one of --add-spincam and --add-tt-spincam may be specified")
		showUsage = True
			
	if showUsage:
		ShowUsage()
		sys.exit(1)

	if pTracker.ImportObjects and not placeAreas:
		unlinkedAreaStartY = 0

	if pTracker.SpectralMode:
		pTracker.IgnoredObjectNames = pTracker.AppendList(pTracker.IgnoredObjectNames, pTracker.MaterialPlaneMonsterObjectNames)
		pTracker.IgnoredObjectNames = pTracker.AppendList(pTracker.IgnoredObjectNames, pTracker.MaterialPlaneOtherObjectNames)
	if pTracker.MaterialMode:
		pTracker.IgnoredObjectNames = pTracker.AppendList(pTracker.IgnoredObjectNames, pTracker.SpectralPlaneMonsterObjectNames)
		pTracker.IgnoredObjectNames = pTracker.AppendList(pTracker.IgnoredObjectNames, pTracker.SpectralPlaneOtherObjectNames)
	
	for whitelisted in pTracker.ObjectNameWhitelist:
		if whitelisted in pTracker.IgnoredObjectNames:
			pTracker.IgnoredObjectNames.remove(whitelisted)
	
	pTracker.AreaJSONDirectory = os.path.dirname(inputFilePath)
	pTracker.AreaColladaDirectory = areasColladaDirectoryPath
	pTracker.ObjectColladaDirectory = objectsColladaDirectoryPath

	if portalOverrideList == "sr1alpha19990123":
		pTracker.PortalOverrides[pTracker.GetAreaToAreaString("nighta2","nighta3")] = 1
		pTracker.PortalOverrides[pTracker.GetAreaToAreaString("nighta3","nighta2")] = 1
		pTracker.PortalOverrides[pTracker.GetAreaToAreaString("tompil2","tompil3")] = 1
		pTracker.PortalOverrides[pTracker.GetAreaToAreaString("tompil3","tompil2")] = 1
		pTracker.PortalOverrides[pTracker.GetAreaToAreaString("undrct1","undrct4")] = 1	
		pTracker.PortalOverrides[pTracker.GetAreaToAreaString("undrct4","undrct1")] = 1	
		pTracker.PortalOverrides[pTracker.GetAreaToAreaString("undrct1","undrct15")] = 1	
		pTracker.PortalOverrides[pTracker.GetAreaToAreaString("undrct15","undrct1")] = 1
		pTracker.PortalOverrides[pTracker.GetAreaToAreaString("undrct8","undrct9")] = 1	
		pTracker.PortalOverrides[pTracker.GetAreaToAreaString("undrct9","undrct8")] = 1	
	if portalOverrideList == "sr1alpha19990204":
		pTracker.PortalOverrides[pTracker.GetAreaToAreaString("nighta2","nighta3")] = 1
		pTracker.PortalOverrides[pTracker.GetAreaToAreaString("nighta3","nighta2")] = 1
		pTracker.PortalOverrides[pTracker.GetAreaToAreaString("tompil2","tompil3")] = 1
		pTracker.PortalOverrides[pTracker.GetAreaToAreaString("tompil3","tompil2")] = 1
		pTracker.PortalOverrides[pTracker.GetAreaToAreaString("undrct1","undrct4")] = 1	
		pTracker.PortalOverrides[pTracker.GetAreaToAreaString("undrct4","undrct1")] = 1	
		pTracker.PortalOverrides[pTracker.GetAreaToAreaString("undrct1","undrct15")] = 1	
		pTracker.PortalOverrides[pTracker.GetAreaToAreaString("undrct15","undrct1")] = 1	
		pTracker.PortalOverrides[pTracker.GetAreaToAreaString("undrct8","undrct9")] = 1	
		pTracker.PortalOverrides[pTracker.GetAreaToAreaString("undrct9","undrct8")] = 1	
	if portalOverrideList == "sr1alpha19990216":
		pTracker.PortalOverrides[pTracker.GetAreaToAreaString("nighta2","nighta3")] = 1
		pTracker.PortalOverrides[pTracker.GetAreaToAreaString("nighta3","nighta2")] = 1
		pTracker.PortalOverrides[pTracker.GetAreaToAreaString("tompil2","tompil3")] = 1
		pTracker.PortalOverrides[pTracker.GetAreaToAreaString("tompil3","tompil2")] = 1
		pTracker.PortalOverrides[pTracker.GetAreaToAreaString("undrct1","undrct4")] = 1	#
		pTracker.PortalOverrides[pTracker.GetAreaToAreaString("undrct4","undrct1")] = 1	#
		pTracker.PortalOverrides[pTracker.GetAreaToAreaString("undrct1","undrct15")] = 1	#
		pTracker.PortalOverrides[pTracker.GetAreaToAreaString("undrct15","undrct1")] = 1	#
		pTracker.PortalOverrides[pTracker.GetAreaToAreaString("undrct8","undrct9")] = 1	#
		pTracker.PortalOverrides[pTracker.GetAreaToAreaString("undrct9","undrct8")] = 1	#
	if portalOverrideList == "sr1beta19990512":
		pTracker.PortalOverrides[pTracker.GetAreaToAreaString("cathy16","cathy70")] = 2	# correct
		pTracker.PortalOverrides[pTracker.GetAreaToAreaString("cathy70","cathy16")] = 2
		pTracker.PortalOverrides[pTracker.GetAreaToAreaString("nighta2","nighta3")] = 1
		pTracker.PortalOverrides[pTracker.GetAreaToAreaString("nighta3","nighta2")] = 1
		pTracker.PortalOverrides[pTracker.GetAreaToAreaString("tompil2","tompil3")] = 1
		pTracker.PortalOverrides[pTracker.GetAreaToAreaString("tompil3","tompil2")] = 1
		pTracker.BlacklistedPortalCombinations.append("cathy16,119:cathy15,120")	# wrong vertical offset, completely messes up the top of the Cathedral
																					# might be due to use of 16-bit integers and wrapping around?
	if portalOverrideList == "sr1ps1ntsc":
		pTracker.PortalOverrides[pTracker.GetAreaToAreaString("cathy17","cathy49")] = 1
		pTracker.PortalOverrides[pTracker.GetAreaToAreaString("cathy49","cathy17")] = 1
		pTracker.PortalOverrides[pTracker.GetAreaToAreaString("cathy15","cathy49")] = 1
		pTracker.PortalOverrides[pTracker.GetAreaToAreaString("cathy49","cathy15")] = 1
		pTracker.PortalOverrides[pTracker.GetAreaToAreaString("nighta2","nighta3")] = 1
		pTracker.PortalOverrides[pTracker.GetAreaToAreaString("nighta3","nighta2")] = 1
		pTracker.PortalOverrides[pTracker.GetAreaToAreaString("tompil2","tompil3")] = 1
		pTracker.PortalOverrides[pTracker.GetAreaToAreaString("tompil3","tompil2")] = 1

	# get a list of all of the area files, so that unlinked areas can be imported separately
	if placeAreas:
		for areaColladaFile in os.listdir(areasColladaDirectoryPath):
			if areaColladaFile.endswith(colladaExtension):
				currentAreaColladaFilePath = os.path.join(areasColladaDirectoryPath, areaColladaFile)
				#currentAreaColladaFilePath = currentAreaColladaFilePath.replace('\\', '\\\\')
				pTracker.AllAreaColladaFiles.append(currentAreaColladaFilePath)
	# if an objects directory was specified, get a list of all of the Collada files
	if pTracker.ImportObjects:
		if objectsColladaDirectoryPath != "":
				for objectColladaFile in os.listdir(objectsColladaDirectoryPath):
					if objectColladaFile.endswith(colladaExtension):
						currentObjectColladaFilePath = os.path.join(objectsColladaDirectoryPath, objectColladaFile)
						#currentObjectColladaFilePath = currentObjectColladaFilePath.replace('\\', '\\\\')
						pTracker.AllObjectColladaFiles.append(currentObjectColladaFilePath)
	
	# Header and functions
	outputFile = open(outputFilePath,"w")
	scriptFunctions = """#!/usr/bin/env python3
	
import bpy
import math
import numpy 
import re
import time

gWriteToLogFile = True
gLineSeparator = "\\n"
vertexColourAttributeName = "VertexColour"

def WriteToFile(outputFilePath, content):
	global gWriteToLogFile
	print(content)
	if gWriteToLogFile and outputFilePath != "":
		try:
			outputFile = open(outputFilePath, "a", newline=gLineSeparator)
			outputFile.write(content)
			outputFile.write(gLineSeparator)
			outputFile.close()
		except Exception as e:
			gWriteToLogFile = False
			print("Error writing to file '%s': %s. Logging will be disabled." % (outputFilePath, e))

def getNodeByName(material, nodeName):
	for n in material.node_tree.nodes:
		if n.name == nodeName:
			return n
	return None

# https://blender.stackexchange.com/questions/151319/adding-camera-to-scene
def getNewCamera(name, focalLength = 50):
	newCam = bpy.data.cameras.new(name)
	newCam.lens = focalLength
	newCamObject = bpy.data.objects.new(name, newCam)
	newCamObject.location = (0, 0, 0)
	newCamObject.rotation_euler = (0, 0, 0)
	bpy.context.scene.collection.objects.link(newCamObject)
	return newCamObject

def getNewCube():
    existing_objects = set(bpy.context.scene.objects)
    bpy.ops.mesh.primitive_cube_add(size=1, enter_editmode=False, location=(0, 0, 0))
    new_cube_collection = set(bpy.context.scene.objects) - existing_objects
    result = None
    for nc in new_cube_collection:
        result = nc
    return result

def getNewCircle():
    existing_objects = set(bpy.context.scene.objects)
    bpy.ops.curve.primitive_bezier_circle_add(radius=10, enter_editmode=False, location=(0, 0, 0))
    new_obj_collection = set(bpy.context.scene.objects) - existing_objects
    result = None
    for nc in new_obj_collection:
        result = nc
    return result

# https://stackoverflow.com/questions/17355617/can-you-add-a-light-source-in-blender-using-python
def getNewLight(lightName, lightType, lightStrength = 5.0):
	# create light datablock, set attributes
	ldb = bpy.data.lights.new(name = lightName, type = lightType)
	ldb.energy = lightStrength

	# create new object with our light datablock
	new_light = bpy.data.objects.new(name = lightName, object_data = ldb)

	# link light object
	bpy.context.collection.objects.link(new_light)

	return new_light

# https://stackoverflow.com/questions/33186740/fitting-exponential-function-through-two-data-points-with-scipy-curve-fit


# have to do this by hand using math, because Blender doesn't seem to support creating a Bezier circle and then updating its radius
def doSpinCam(cam, targetObject, numRotations, startRadius, endRadius, startZ, endZ, startFrame, endFrame, useExponentialSpeedForSpin = False, useExponentialSpeedForDistance = False, useExponentialSpeedForHeight = False, speedExponent = 2.0, debugLogPath = ""):
    tcon = cam.constraints.new(type='TRACK_TO')
    tcon.target = targetObject
    tcon.track_axis = 'TRACK_NEGATIVE_Z'
    tcon.up_axis = 'UP_Y'
    frameCount = endFrame - startFrame + 1
    rotationSteps = []
    radiusSteps = []
    zSteps = []
    exponentialSegments = []
    xStep = 1.0 / float(frameCount)
    currentX = 0
	camRotation = (math.pi / 2) * -1.0
    for i in range(0, frameCount):
        currentY = math.pow(float(xStep) * float(i), float(speedExponent))
        exponentialSegments.append(currentY)
        #print("X = %s, Y = %s" % (i, currentY))
        currentX += xStep
    if useExponentialSpeedForSpin:
        for i in range(0, frameCount):
            rotationSteps.append((float(numRotations) * 2.0 * math.pi) * exponentialSegments[i])
    else:
        rotationStep = (float(numRotations) * 2.0 * math.pi) / float(frameCount)
        for i in range(0, frameCount):
			camRotation += rotationStep
			rotationSteps.append(camRotation)
            #rotationSteps.append(rotationStep * float(i))
    if useExponentialSpeedForDistance:
        for i in range(0, frameCount):
            radiusSteps.append((float(endRadius) - float(startRadius)) * exponentialSegments[i])
    else:
        radiusStep = (float(endRadius) - float(startRadius)) / float(frameCount)
        for i in range(0, frameCount):
            radiusSteps.append(radiusStep * float(i))
    if useExponentialSpeedForHeight:
        for i in range(0, frameCount):
            zSteps.append((float(endZ) - float(startZ)) * exponentialSegments[i])
    else:
        zStep = (float(endZ) - float(startZ)) / float(frameCount)
        for i in range(0, frameCount):
            zSteps.append(zStep * float(i))
    for frameNum in range(0, frameCount):
        absFrameNum = frameNum + startFrame
        bpy.context.scene.frame_set(absFrameNum)
        currentRadius = startRadius + (radiusSteps[frameNum])
        camLocationX = currentRadius * math.cos(rotationSteps[frameNum])
        camLocationY = currentRadius * math.sin(rotationSteps[frameNum])
        camLocationZ = (startZ + (zSteps[frameNum]))
        #print("Setting spincam position to %s, %s, %s at frame %s/%s" % (camLocationX, camLocationY, camLocationZ, frameNum + 1, frameCount))
        cam.location = (camLocationX, camLocationY, camLocationZ)
        cam.keyframe_insert(data_path="location", index=-1)
        if bpy.context.scene.frame_end < absFrameNum:
            bpy.context.scene.frame_end = absFrameNum


def doTierSpin(cam, spinRadius, startRotation, tierZ, startFrame, rotationFrameCountPerTier, debugLogPath = ""):
	twoPi = 2.0 * math.pi
	rotationStep = twoPi / float(rotationFrameCountPerTier)
	currentRotation = startRotation
	for frameNum in range(0, rotationFrameCountPerTier):
        bpy.context.scene.frame_set(startFrame + frameNum)
        camLocationX = spinRadius * math.cos(currentRotation)
        camLocationY = spinRadius * math.sin(currentRotation)
        camLocationZ = tierZ
        #print("Setting spincam position to %s, %s, %s at frame %s" % (camLocationX, camLocationY, camLocationZ, absFrameNum))
        cam.location = (camLocationX, camLocationY, camLocationZ)
        cam.keyframe_insert(data_path="location", index=-1)
		currentRotation += rotationStep
	return currentRotation
	
def doTransitionSpin(cam, spinRadiusTierA, spinRadiusTierB, startRotation, rotationChange, tierAZ, tierBZ, startFrame, transitionFrameCount, debugLogPath = ""):
	rotationStep = rotationChange / float(transitionFrameCount)
	heightStep = (float(tierAZ) - float(tierBZ)) / float(transitionFrameCount)
	distanceStep = (float(spinRadiusTierA) - float(spinRadiusTierB)) / float(transitionFrameCount)
	currentRotation = startRotation
	currentDistance = spinRadiusTierA
	camLocationZ = tierAZ
	for frameNum in range(0, transitionFrameCount):
        bpy.context.scene.frame_set(startFrame + frameNum)
        camLocationX = currentDistance * math.cos(currentRotation)
        camLocationY = currentDistance * math.sin(currentRotation)
        #print("Setting spincam position to %s, %s, %s at frame %s" % (camLocationX, camLocationY, camLocationZ, absFrameNum))
        cam.location = (camLocationX, camLocationY, camLocationZ)
        cam.keyframe_insert(data_path="location", index=-1)
		currentRotation += rotationStep
		currentDistance -= distanceStep
        camLocationZ -= heightStep
	return currentRotation
		

def doTripleTierSpinCam(cam, targetObject, spinRadiusTier1, spinRadiusTier2, spinRadiusTier3, tier1Z, tier2Z, tier3Z, startFrame, endFrame, debugLogPath = ""):
    tcon = cam.constraints.new(type='TRACK_TO')
    tcon.target = targetObject
    tcon.track_axis = 'TRACK_NEGATIVE_Z'
    tcon.up_axis = 'UP_Y'
    frameCount = endFrame - startFrame + 1
	# frameCount needs to be evenly divisible by eight for some of these integer calculations to work
	previousFrameCount = frameCount
	while frameCount % 8 > 0:
		frameCount += 1 
	if previousFrameCount != frameCount:
		WriteToFile(debugLogPath, "Warning: adjusted frame count of triple spin cam from %s to %s so that it would be evenly-divisible by eight" % (previousFrameCount, frameCount))
	rotationFrameCountPerTier = int(frameCount / 4)
	transitionFrameCount = int(rotationFrameCountPerTier / 4)
	absFrameNum = startFrame
	camRotation = (math.pi / 2) * -1.0
    # tier 1
    camRotation = doTierSpin(cam, spinRadiusTier1, camRotation, tier1Z, absFrameNum, rotationFrameCountPerTier, debugLogPath = debugLogPath)
	absFrameNum += rotationFrameCountPerTier
	# transition 1 to 2
	camRotation = doTransitionSpin(cam, spinRadiusTier1, spinRadiusTier2, camRotation, (math.pi / 2.0), tier1Z, tier2Z, absFrameNum, transitionFrameCount, debugLogPath = debugLogPath)
	absFrameNum += transitionFrameCount
	# tier 2
    camRotation = doTierSpin(cam, spinRadiusTier2, camRotation, tier2Z, absFrameNum, rotationFrameCountPerTier, debugLogPath = debugLogPath)
	absFrameNum += rotationFrameCountPerTier
	# transition 2 to 3
	camRotation = doTransitionSpin(cam, spinRadiusTier2, spinRadiusTier3, camRotation, (math.pi / 2.0), tier2Z, tier3Z, absFrameNum, transitionFrameCount, debugLogPath = debugLogPath)
	absFrameNum += transitionFrameCount
	# tier 3
    camRotation = doTierSpin(cam, spinRadiusTier3, camRotation, tier3Z, absFrameNum, rotationFrameCountPerTier, debugLogPath = debugLogPath)
	absFrameNum += rotationFrameCountPerTier
	# transition 3 to 1
	camRotation = doTransitionSpin(cam, spinRadiusTier3, spinRadiusTier1, camRotation, math.pi, tier3Z, tier1Z, absFrameNum, transitionFrameCount * 2, debugLogPath = debugLogPath)
	absFrameNum += transitionFrameCount * 2
	
	if bpy.context.scene.frame_end < absFrameNum:
		bpy.context.scene.frame_end = absFrameNum

	
def addOverheadLightArray(minX, minY, maxX, maxY, stepX, stepY, zPosition, lightStrength, lightColour, debugLogPath = ""):
	currentX = minX
	currentY = minY
	doneRows = False
	lightNum = 0
	maxCol = 0
	maxRow = 0
	while not doneRows:
		doneCols = False
		maxCol = 0
		while not doneCols:
			lightNum += 1
			newLight = getNewLight("overhead_%s" % (lightNum), 'SUN', lightStrength = lightStrength)
			newLight.location[0] = currentX
			newLight.location[1] = currentY
			newLight.location[2] = zPosition
			newLight.data.color = lightColour
			currentX += stepX
			if currentX > maxX:
				doneCols = True
			maxCol += 1
		currentY += stepY
		currentX = minX
		if currentY > maxY:
			doneRows = True
		maxRow += 1
	WriteToFile(debugLogPath, "Added %s light(s) in %s rows, %s columns" % (lightNum, maxRow, maxCol))
	bpy.context.evaluated_depsgraph_get().update()


def getMostRecentNodeByName(material, nodeNamePrefix):
	result = None
	for n in material.node_tree.nodes:
		if n.name.startswith(nodeNamePrefix):
			result = n
	return result

def getObjectWithShortestName(objectList):
	minLength = 10000
	for o in objectList:
		if len(o.name) < minLength:
			minLength = len(o.name)
	for o in objectList:
		if len(o.name) == minLength:
			return o
	return None

def getImportedTopLevelObjects(objectList):
	result = []
	for o in objectList:
		if not o.parent or o.parent == '':
			result.append(o)
	return result

def renameChildObjects(objectList, oldString, newString):
	for o in objectList:
		if o.name != oldString and oldString in o.name:
			o.name = o.name.replace(oldString, newString)
			if o.data:
				o.data.name = o.data.name.replace(oldString, newString)

# loosely based on https://blenderartists.org/t/question-regarding-expanding-collapsing-collection-in-outliner-in-2-8/1175242
# the original version of the code only expanded/collapsed the outliner in the Scripting view
# this version applies the changes to all views
def toggle_expand(state):
	for s in bpy.data.screens.keys():
		bpy.ops.wm.redraw_timer(type='DRAW_WIN', iterations=1)
		for a in bpy.data.screens[s].areas:
			if a.type == 'OUTLINER':
				bpy.ops.outliner.show_hierarchy({'area': a}, 'INVOKE_DEFAULT')
				for i in range(state):
					bpy.ops.outliner.expanded_toggle({'area': a})
				a.tag_redraw()

def setMaterialDefaultAlphaTo1(mat, debugLogPath = ""):
	bsdfNode = None
	for n in mat.node_tree.nodes:
		if n.name == 'Principled BSDF':
			bsdfNode = n
		if n.name == 'Image Texture':
			baseColourNode = n
		if n.name == 'Vertex Color':
			vertexColourNode = n
		if n.name == 'Mix':
			mixShaderNode = n
		if n.name == 'Value':
			valueNode = n
		if n.name == 'Material Output':
			materialOutputNode = n
	if bsdfNode:
		if bsdfNode.inputs['Alpha'].default_value == 0.0:
			bsdfNode.inputs['Alpha'].default_value = 1.0

def setAllMaterialsDefaultAlphaTo1(debugLogPath = ""):
	WriteToFile(debugLogPath, "Setting default alpha for all materials to 1.0 if it is 0.0")
	materialNum = 1
	materialCount = len(bpy.data.materials)
	materialList = []
	for mat in bpy.data.materials:
		materialList.append(mat.name)
	for i in range(0, len(materialList)):
		setMaterialDefaultAlphaTo1(bpy.data.materials[materialList[i]], debugLogPath = debugLogPath)
		materialNum += 1
		if materialNum == materialCount or (materialNum % 100) == 0:
			WriteToFile(debugLogPath, "%s\t%s/%s materials processed" % (time.time(), materialNum, materialCount))
		
def adjustMaterial(mat, mixFactor = 1.0, useSpecialMaterialCases = True, everythingGlows = False, brightVertexColours = False, debugLogPath = "", emulateBackfaceCulling = True):
	processMaterial = True
	opacity = 1.0
	emissivity = 0.0
	# Ignore Blender's default material
	if mat.name == "Material":
		processMaterial = False
	if processMaterial:
		# enable backface culling (single-sided polygons)
		#WriteToFile(debugLogPath, "Material: %s" % mat.name)
		if '_op_' in mat.name:
			#opacity_string = re.search('op_[0-9\.,_]+_$', mat.name).group()
			regex_results = re.search('op_[0-9_]+_($|.)', mat.name)
			if regex_results:
				opacity_string = regex_results.group()
				if opacity_string != '':
					# enable alpha blend mode (for transparent/translucent textures)
					opacity_string = opacity_string[3:]
					opacity_string = opacity_string[:-1]
					opacity_string = opacity_string.replace("_", ".")
					while (opacity_string.endswith(".")):
						opacity_string = opacity_string[:-1]
					try:
						opacity = float(opacity_string)
					except Exception as e:
						WriteToFile(debugLogPath, "Couldn't convert '%s' to a floating-point number" % (opacity_string))
						# handle locales that use comma as a decimal point instead of a period
						opacity_string = opacity_string.replace(",", ".")
						try:
							opacity = float(opacity_string)
						except Exception as e:
							WriteToFile(debugLogPath, "Couldn't convert '%s' to a floating-point number" % (opacity_string))
					if opacity < 1.0:
						mat.blend_method = 'BLEND'
		else:
			if "_am" in mat.name:
				# enable alpha clipping mode (for opaque extures with some 100% transparent parts)
				mat.blend_method = 'CLIP'
		if '_gl_' in mat.name:
			regex_results = re.search('gl_[0-9_]+_($|.)', mat.name)
			if regex_results:
				emissivity_string = regex_results.group()
				if emissivity_string != '':
					# enable alpha blend mode (for transparent/translucent textures)
					emissivity_string = emissivity_string[3:]
					emissivity_string = emissivity_string[:-1]
					emissivity_string = emissivity_string.replace("_", ".")
					while (emissivity_string.endswith(".")):
						emissivity_string = emissivity_string[:-1]
					try:
						emissivity = float(emissivity_string)
					except Exception as e:
						WriteToFile(debugLogPath, "Couldn't convert '%s' to a floating-point number" % (emissivity_string))
						# handle locales that use comma as a decimal point instead of a period
						emissivity_string = emissivity_string.replace(",", ".")
						try:
							emissivity = float(emissivity_string)
						except Exception as e:
							WriteToFile(debugLogPath, "Couldn't convert '%s' to a floating-point number" % (emissivity_string))
		# link texture alpha to BSDF alpha and create a custom shader pipeline to emulate the PlayStation
		mat.use_nodes= True
		bsdfNode = None
		baseColourNode = None
		mixShaderNode = None
		vertexColourNode = None
		valueNode = None
		separateRGBNode = None
		combineRGBNode = None
		multiplyNodeR = None
		multiplyNodeG = None
		multiplyNodeB = None
		materialOutputNode = None
		materialElementType = "unknown"
		bfceTransparentBSDFNode = None
		bfceGeometryNode = None
		bfceLightPathNode = None
		bfceMixShaderNode1 = None
		bfceMixShaderNode2 = None			
		if re.match("^m[0-9]+o_", mat.name):
			materialElementType = "object"
		if re.match("^m[0-9]+a_", mat.name):
			materialElementType = "area"
		if re.match("^m[0-9]+d_", mat.name):
			materialElementType = "debug"
		#WriteToFile(debugLogPath, "Material '%s' type is '%s'" % (mat.name, materialElementType))
		multiplicationFactor = 4.0	# "Modulate 4x"
		for n in mat.node_tree.nodes:
			#WriteToFile(debugLogPath, n.name)
			if n.name == 'Principled BSDF':
				bsdfNode = n
			if n.name == 'Image Texture':
				baseColourNode = n
			if n.name == 'Vertex Color':
				vertexColourNode = n
			if n.name == 'Mix':
				mixShaderNode = n
			if n.name == 'Value':
				valueNode = n
			if n.name == 'Material Output':
				materialOutputNode = n
		if bsdfNode and not baseColourNode:
			# importing a Collada file with only a polygon colour defined uses the colour value in the BSDF node 
			# instead of having a separate one. That won't work for this purpose, so create a new one with a clone
			# of the BSDF node's colour
			mat.node_tree.nodes.new(type = "ShaderNodeRGB")
			baseColourNode = getMostRecentNodeByName(mat, "RGB")
			baseColourNode.outputs[0].default_value[0] = bsdfNode.inputs[0].default_value[0]
			baseColourNode.outputs[0].default_value[1] = bsdfNode.inputs[0].default_value[1]
			baseColourNode.outputs[0].default_value[2] = bsdfNode.inputs[0].default_value[2]
			baseColourNode.outputs[0].default_value[3] = bsdfNode.inputs[0].default_value[3]
		if bsdfNode and baseColourNode:
			if not mixShaderNode or not vertexColourNode:
				if opacity < 1.0:
					mat.node_tree.nodes.new(type = "ShaderNodeValue")
					valueNode = getMostRecentNodeByName(mat, "Value")
					valueNode.outputs[0].default_value = opacity
					mat.node_tree.links.new(valueNode.outputs[0], bsdfNode.inputs['Alpha'])
				else:
					if 'Alpha' in baseColourNode.outputs.keys():
						mat.node_tree.links.new(baseColourNode.outputs['Alpha'], bsdfNode.inputs['Alpha'])
					else:
						bsdfNode.inputs['Alpha'].default_value = baseColourNode.outputs[0].default_value[3]
				mat.node_tree.nodes.new(type = "ShaderNodeMixRGB")
				mixShaderNode = getMostRecentNodeByName(mat, "Mix")
				mixShaderNode.inputs['Fac'].default_value = mixFactor
				mixShaderNode.blend_type = 'MULTIPLY'
				if brightVertexColours:
					mixShaderNode.blend_type = 'COLOR'
				
				# areas use their actual vertex colours
				# objects use a 100% white vertex colour which is set based on the closest
				# terrain vertex colour later. I think this is more or less how the game works, 
				# since it seems like a decent way to fake lighting effects on older hardware,
				# but in any case, it works reasonably well
				if materialElementType == "area" or materialElementType == "debug":
					# Vertex colour shader node doesn't work correctly in Cycles or the Material Preview
					#mat.node_tree.nodes.new(type = "ShaderNodeVertexColor")
					#vertexColourNode = getMostRecentNodeByName(mat, "Vertex Color")
					mat.node_tree.nodes.new(type = "ShaderNodeAttribute")
					vertexColourNode = getMostRecentNodeByName(mat, "Attribute")
					vertexColourNode.attribute_name = vertexColourAttributeName
				else:
					mat.node_tree.nodes.new(type = "ShaderNodeRGB")
					vertexColourNode = getMostRecentNodeByName(mat, "RGB")
					vertexColourNode.outputs[0].default_value[0] = 1.0
					vertexColourNode.outputs[0].default_value[1] = 1.0
					vertexColourNode.outputs[0].default_value[2] = 1.0
					vertexColourNode.outputs[0].default_value[3] = 1.0
				
				mat.node_tree.nodes.new(type = "ShaderNodeMath")
				multiplyNodeR = getMostRecentNodeByName(mat, "Math")
				multiplyNodeR.operation = 'MULTIPLY'
				multiplyNodeR.inputs[1].default_value = multiplicationFactor
				mat.node_tree.nodes.new(type = "ShaderNodeMath")
				multiplyNodeG = getMostRecentNodeByName(mat, "Math")
				multiplyNodeG.operation = 'MULTIPLY'
				multiplyNodeG.inputs[1].default_value = multiplicationFactor
				mat.node_tree.nodes.new(type = "ShaderNodeMath")
				multiplyNodeB = getMostRecentNodeByName(mat, "Math")
				multiplyNodeB.operation = 'MULTIPLY'
				multiplyNodeB.inputs[1].default_value = multiplicationFactor
				
				mat.node_tree.nodes.new(type = "ShaderNodeSeparateRGB")
				separateRGBNode = getMostRecentNodeByName(mat, "Separate RGB")
				
				mat.node_tree.nodes.new(type = "ShaderNodeCombineRGB")
				combineRGBNode = getMostRecentNodeByName(mat, "Combine RGB")
				
				
				# Remove existing links, where necessary
				for existingLink in baseColourNode.outputs['Color'].links:
					mat.node_tree.links.remove(existingLink)
				if brightVertexColours:
					mat.node_tree.links.new(baseColourNode.outputs['Color'], mixShaderNode.inputs['Color1'])
				else:
					mat.node_tree.links.new(baseColourNode.outputs['Color'], separateRGBNode.inputs[0])
					mat.node_tree.links.new(separateRGBNode.outputs[0], multiplyNodeR.inputs[0])
					mat.node_tree.links.new(separateRGBNode.outputs[1], multiplyNodeG.inputs[0])
					mat.node_tree.links.new(separateRGBNode.outputs[2], multiplyNodeB.inputs[0])
					mat.node_tree.links.new(multiplyNodeR.outputs[0], combineRGBNode.inputs[0])
					mat.node_tree.links.new(multiplyNodeG.outputs[0], combineRGBNode.inputs[1])
					mat.node_tree.links.new(multiplyNodeB.outputs[0], combineRGBNode.inputs[2])
					mat.node_tree.links.new(combineRGBNode.outputs['Image'], mixShaderNode.inputs['Color1'])
				mat.node_tree.links.new(vertexColourNode.outputs['Color'], mixShaderNode.inputs['Color2'])
				mat.node_tree.links.new(mixShaderNode.outputs['Color'], bsdfNode.inputs['Base Color'])
				# set specular to 0 to more accurately reflect the game
				bsdfNode.inputs['Specular'].default_value = 0.0
				# if the material is glowing, attach the mixed colour output to the glow (Emission) input
				if emissivity > 0.0:
					emissivityHandled = False
					if useSpecialMaterialCases:
						redEyes = False
						
						if "o_aluka_" in mat.name:
							redEyes = True
						if "o_morlock_" in mat.name:
							redEyes = True
						if "o_ronin_" in mat.name:
							redEyes = True
						if "o_skinner_" in mat.name:
							redEyes = True
						if "o_wallcr_" in mat.name:
							redEyes = True
						if redEyes:
							WriteToFile(debugLogPath, "Using special material handling for %s - enhanced vampire eye glow" % (mat.name))
							mat.node_tree.nodes.new(type = "ShaderNodeRGB")
							glowNode = getMostRecentNodeByName(mat, "RGB")
							glowNode.outputs[0].default_value[0] = 1.0
							glowNode.outputs[0].default_value[1] = 0.0
							glowNode.outputs[0].default_value[2] = 0.0
							glowNode.outputs[0].default_value[3] = 1.0
							mat.node_tree.links.new(glowNode.outputs['Color'], bsdfNode.inputs['Emission'])
							emissivityHandled = True
					if not emissivityHandled:
						mat.node_tree.links.new(mixShaderNode.outputs['Color'], bsdfNode.inputs['Emission'])
				if everythingGlows:
					try:
						mat.node_tree.links.new(mixShaderNode.outputs['Color'], bsdfNode.inputs['Emission'])
					except Exception as e:
						WriteToFile(debugLogPath, "Couldn't link mixShaderNode to emission for material '%s': %s" % (mat.name, e))
				# delete the Principled BSDF => Material Output connection and replace it with a really complicated
				# set of additional nodes, because Cycles doesn't have a built-in feature to cull backfacing polygons
				# ...but only do this if the material is not transparent, glowing, or alphamasked, because that looks wrong too
				emulateBFC = emulateBackfaceCulling
				if opacity < 1.0:
					emulateBFC = False
				#if emissivity > 0.0:
				#	emulateBFC = False
				if mat.blend_method == 'CLIP':
					emulateBFC = False
				if mat.blend_method == 'BLEND':
					emulateBFC = False
				if emulateBFC:
					# enable actual backface culling for renderers that support it
					mat.use_backface_culling = True
					# disable display of transparent backs, because it causes visual corruption when backface culling is enabled
					mat.show_transparent_back = False
					
					for existingLink in bsdfNode.outputs[0].links:
						mat.node_tree.links.remove(existingLink)
						
					mat.node_tree.nodes.new(type = "ShaderNodeBsdfTransparent")
					bfceTransparentBSDFNode = getMostRecentNodeByName(mat, "Transparent")
					
					mat.node_tree.nodes.new(type = "ShaderNodeNewGeometry")
					bfceGeometryNode = getMostRecentNodeByName(mat, "Geometry")
					
					mat.node_tree.nodes.new(type = "ShaderNodeLightPath")
					bfceLightPathNode = getMostRecentNodeByName(mat, "Light Path")
					
					mat.node_tree.nodes.new(type = "ShaderNodeMixShader")
					bfceMixShaderNode1 = getMostRecentNodeByName(mat, "Mix")

					mat.node_tree.nodes.new(type = "ShaderNodeMixShader")
					bfceMixShaderNode2 = getMostRecentNodeByName(mat, "Mix")
					
					mat.node_tree.links.new(bfceGeometryNode.outputs['Backfacing'], bfceMixShaderNode1.inputs['Fac'])
					mat.node_tree.links.new(bsdfNode.outputs['BSDF'], bfceMixShaderNode1.inputs[1])
					mat.node_tree.links.new(bfceTransparentBSDFNode.outputs['BSDF'], bfceMixShaderNode1.inputs[2])
					
					mat.node_tree.links.new(bfceLightPathNode.outputs['Is Shadow Ray'], bfceMixShaderNode2.inputs['Fac'])
					mat.node_tree.links.new(bfceMixShaderNode1.outputs['Shader'], bfceMixShaderNode2.inputs[1])
					mat.node_tree.links.new(bfceTransparentBSDFNode.outputs['BSDF'], bfceMixShaderNode2.inputs[2])
					
					mat.node_tree.links.new(bfceMixShaderNode2.outputs['Shader'], materialOutputNode.inputs['Surface'])
					
			else:
				WriteToFile(debugLogPath, "mixShaderNode and vertexColourNode already exist in material %s" % (mat.name))
		else:
			WriteToFile(debugLogPath, "Material %s is missing bsdfNode or baseColourNode" % (mat.name))

def adjustMaterials(mixFactor = 1.0, useSpecialMaterialCases = True, everythingGlows = False, brightVertexColours = False, debugLogPath = "", emulateBackfaceCulling = True):
	WriteToFile(debugLogPath, "Adjusting materials")
	materialNum = 1
	materialCount = len(bpy.data.materials)
	materialList = []
	for mat in bpy.data.materials:
		materialList.append(mat.name)
	for i in range(0, len(materialList)):
		adjustMaterial(bpy.data.materials[materialList[i]], mixFactor = mixFactor, useSpecialMaterialCases = useSpecialMaterialCases, everythingGlows = everythingGlows, brightVertexColours = brightVertexColours, debugLogPath = debugLogPath, emulateBackfaceCulling = emulateBackfaceCulling)
		materialNum += 1
		if materialNum == materialCount or (materialNum % 100) == 0:
			WriteToFile(debugLogPath, "%s\t%s/%s materials processed" % (time.time(), materialNum, materialCount))

def deleteEmptyObject(obj, debugLogPath = ""):
	try:
		mesh = obj.data
		
		if not mesh:
			deleteObject = True
			if obj.name.startswith("area_"):
				deleteObject = False
			if obj.name.startswith("awo_"):
				deleteObject = False
			if obj.name.startswith("aset_"):
				deleteObject = False
			if "_obj_" in obj.name:
				deleteObject = False
			if deleteObject:
				WriteToFile(debugLogPath, "Object %s was missing the data attribute - deleting this object" % (obj.name))
				bpy.ops.object.select_all(action='DESELECT')
				obj.select_set(True)
				bpy.ops.object.delete()
				bpy.ops.object.select_all(action='DESELECT')
	except Exception as e:
		WriteToFile(debugLogPath, "Object %s high-level exception while deleting empty objects: %s" % (obj.name, e))

def deleteEmptyObjects(debugLogPath = ""):
	WriteToFile(debugLogPath, "Deleting empty objects")
	eNum = 0
	objList = []
	for obj in bpy.data.objects:
		objList.append(obj.name)
	for i in range(0, len(objList)):
		deleteEmptyObject(bpy.data.objects[objList[i]], debugLogPath = debugLogPath)
		eNum += 1
		if eNum == len(objList) or (eNum % 100) == 0:
			msg = "%s\t%s/%s objects processed" % (time.time(), eNum, len(objList))
			WriteToFile(debugLogPath, msg)
		

def createMissingVertexColoursForObject(obj, debugLogPath = ""):
	try:
		mesh = obj.data
		if mesh:
			try:
				if not mesh.vertex_colors:
					WriteToFile(debugLogPath, "Object %s was missing the vertex_colors layer" % (obj.name))
					mesh.vertex_colors.new()
				color_layer = mesh.vertex_colors.active
				# Rename vertex colour attributes globally so that they work correctly with the 
				# shader workaround mentioned in adjustMaterials
				color_layer.name = vertexColourAttributeName
				i = 0
				for poly in mesh.polygons:
					for idx in poly.loop_indices:
						setColour = False
						if len(color_layer.data) < (i + 1):
							setColour = True
						if not color_layer.data[i]:
							setColour = True
						if setColour:
							color_map.data[i].color = (1.0, 1.0, 1.0, 1.0)
							WriteToFile(debugLogPath, "\tObject %s - set vertex colour #%s" % (obj.name, i))
						i += 1
				#WriteToFile(debugLogPath, "\tObject %s: %s vertex colours" % (obj.name, i))
			except Exception as e:
				WriteToFile(debugLogPath, "Object %s exception while creating missing vertex colours: %s" % (obj.name, e))
	except Exception as e:
		WriteToFile(debugLogPath, "Object %s high-level exception while creating missing vertex colours: %s" % (obj.name, e))

def createMissingVertexColours(debugLogPath = ""):
	WriteToFile(debugLogPath, "Creating missing vertex colour layers")
	eNum = 0
	objList = []
	for obj in bpy.data.objects:
		objList.append(obj.name)
	for i in range(0, len(objList)):
		createMissingVertexColoursForObject(bpy.data.objects[objList[i]], debugLogPath = debugLogPath)
		eNum += 1
		if eNum == len(objList) or (eNum % 100) == 0:
			msg = "%s\t%s/%s objects processed" % (time.time(), eNum, len(objList))
			WriteToFile(debugLogPath, msg)
		
def materialIsOrphaned(mat, debugLogPath = ""):
	try:
		used_by = []
		for obj in bpy.data.objects:
			try:
				for slot in obj.material_slots:
					try:
						if slot.material == mat:
							used_by.append(obj)
					except Exception as e:
						WriteToFile(debugLogPath, "Object %s exception while checking for orphaned materials in material slots: %s" % (obj.name, e))
			except Exception as e:
				WriteToFile(debugLogPath, "Object %s exception while checking for orphaned materials: %s" % (obj.name, e))
		if len(used_by) == 0:
			WriteToFile(debugLogPath, "Material %s is orphaned, and will be deleted" % (mat.name))
			return True
		return False
	except Exception as e:
		WriteToFile(debugLogPath, "Material %s exception while checking for orphaned materials: %s" % (mat.name, e))

def deleteOrphanedMaterials(debugLogPath = ""):
	WriteToFile(debugLogPath, "Deleting orphaned materials")
	eNum = 0
	#delete_materials = []
	materialList = []
	for mat in bpy.data.materials:
		materialList.append(mat.name)
	for i in range(0, len(materialList)):
		m = bpy.data.materials[materialList[i]]
		if materialIsOrphaned(m, debugLogPath = debugLogPath):
			bpy.data.materials.remove(m)
		eNum += 1
		if eNum == len(materialList) or (eNum % 100) == 0:
			msg = "%s\t%s/%s materials processed" % (time.time(), eNum, len(materialList))
			WriteToFile(debugLogPath, msg)
		
	#for dm in delete_materials:
		#bpy.data.materials.remove(dm)

def arrangeShaderNodes(mat, debugLogPath = "", nodeMinSpacingX = 500.0, nodeMinSpacingY = 400.0):
	if len(mat.node_tree.nodes) > 1:
		#WriteToFile(debugLogPath, "Arranging nodes for material: %s" % (mat.name))
		previousColumnNodes = []
		
		# first, find any nodes without connected inputs - this will be the leftmost column
		for n in mat.node_tree.nodes:
			includeNode = False
			if len(n.inputs) == 0:
				includeNode = True
			if not includeNode:
				hasConnectedInputs = False
				for nInput in n.inputs:
					if len(nInput.links) > 0:
						hasConnectedInputs = True
						break
				if not hasConnectedInputs:
					includeNode = True
			if includeNode:
				previousColumnNodes.append(n)
				#WriteToFile(debugLogPath, "\tNode: %s will be included in the leftmost column" % (n.name))
		startX = previousColumnNodes[0].location.x
		startY = previousColumnNodes[0].location.y
		
		currentX = 0.0
		
		while len(previousColumnNodes) > 0:
			maxWidth = 0.0
			nextColumnNodes = []
			currentY = startY
			for n in previousColumnNodes:
				nodeWidth = n.dimensions.x
				nodeHeight = n.dimensions.y
				maxTries = 2
				tryNum = 0
				#while tryNum < maxTries:
				#	if nodeWidth > 0 and nodeHeight > 0:
				#		break
				#	bpy.ops.wm.redraw_timer(type='DRAW_WIN_SWAP', iterations=1)
				#	tryNum += 1
				#	if tryNum >= maxTries:
				#		if nodeWidth == 0:
				#			nodeWidth = nodeMinSpacingX
				#		if nodeHeight == 0:
				#			nodeHeight = nodeMinSpacingY
				#		break
				if nodeWidth > maxWidth:
					maxWidth = nodeWidth
				n.location.y = currentY
				n.location.x = currentX
				currentY -= nodeHeight + nodeMinSpacingY
				#WriteToFile(debugLogPath, "\tChecking outputs for node %s" % (n.name))
				for nOutput in n.outputs:
					for outputLink in nOutput.links:
						if outputLink.to_node not in nextColumnNodes:
							nextColumnNodes.append(outputLink.to_node)
						#WriteToFile(debugLogPath, "\t\tFrom %s to %s" % (outputLink.from_node, outputLink.to_node))
			currentX += maxWidth + nodeMinSpacingX
			previousColumnNodes = nextColumnNodes			
						
def arrangeAllShaderNodes(debugLogPath = "", nodeMinSpacingX = 200.0, nodeMinSpacingY = 100.0):
	WriteToFile(debugLogPath, "Arranging shader nodes")
	eNum = 0
	materialList = []
	for mat in bpy.data.materials:
		materialList.append(mat.name)
	for i in range(0, len(materialList)):
		eNum += 1
		if eNum == len(materialList) or (eNum % 100) == 0:
			msg = "%s\t%s/%s materials processed" % (time.time(), eNum, len(materialList))
			WriteToFile(debugLogPath, msg)
		arrangeShaderNodes(bpy.data.materials[materialList[i]], debugLogPath = debugLogPath, nodeMinSpacingX = nodeMinSpacingX, nodeMinSpacingY = nodeMinSpacingY)

def getObjectWithShortestName(objectList):
    minLength = 10000
    for o in objectList:
        if len(o.name) < minLength:
            minLength = len(o.name)
    for o in objectList:
        if len(o.name) == minLength:
            return o
    return None
	
def getImportedTopLevelObjects(objectList):
    result = []
    for o in objectList:
        if not o.parent or o.parent == '':
            result.append(o)
    return result
	
def renameChildObjects(objectList, oldString, newString):
    for o in objectList:
        if o.name != oldString and oldString in o.name:
            o.name = o.name.replace(oldString, newString)
            if o.data:
                o.data.name = o.data.name.replace(oldString, newString)

# loosely based on https://blenderartists.org/t/question-regarding-expanding-collapsing-collection-in-outliner-in-2-8/1175242
# the original version of the code only expanded/collapsed the outliner in the Scripting view
# this version applies the changes to all views
def toggle_expand(state):
    for s in bpy.data.screens.keys():
        bpy.ops.wm.redraw_timer(type='DRAW_WIN', iterations=1)
        for a in bpy.data.screens[s].areas:
            if a.type == 'OUTLINER':
                bpy.ops.outliner.show_hierarchy({'area': a}, 'INVOKE_DEFAULT')
                for i in range(state):
                    bpy.ops.outliner.expanded_toggle({'area': a})
                a.tag_redraw()

def importObject(objectBaseName, objectColladaPath, parentAreaName, parentAreaWithObjects, parentAreaObjectNumber, placementOrder, distanceFromOrigin, posX, posY, posZ, rotA, rotB, rotC, debugLogPath = ""):
	#objectName = "%s_obj_%s_%s" % (parentAreaName, parentAreaObjectNumber, objectBaseName)
	objectName = objectBaseName
	try:
		preimport_objects = set(bpy.context.scene.objects)
		bpy.ops.wm.collada_import(filepath=objectColladaPath)
		imported_objects = set(bpy.context.scene.objects) - preimport_objects
		if imported_objects and len(imported_objects) > 0:
			current_object_objects = getImportedTopLevelObjects(imported_objects)
			for current_object_object in current_object_objects:
				current_object_object.location[0] = posX
				current_object_object.location[1] = posY
				current_object_object.location[2] = posZ
				current_object_object.rotation_euler[0] = rotA
				current_object_object.rotation_euler[1] = rotB
				current_object_object.rotation_euler[2] = rotC
				current_object_object['scripted_placement_order'] = placementOrder
				current_object_object['distance_from_origin'] = distanceFromOrigin
				current_object_object['g2_type'] = 'object'
				original_object_tlo_name = current_object_object.name
				current_object_object.name = objectName
				if current_object_object.data:
					current_object_object.data.name = objectName
					renameChildObjects(imported_objects, original_object_tlo_name, current_object_object.name)
	except Exception as e:
		errMessage = 'Error importing "%s": %s' % (objectColladaPath, e)
		WriteToFile(debugLogPath, errMessage)
	if objectName in bpy.data.objects.keys():
		try:
			parentAreaWithObjects.objects.link(bpy.data.objects[objectName])
		except Exception as e:
			errMessage = 'Error linking "%s": %s' % (objectColladaPath, e)
			WriteToFile(debugLogPath, errMessage)

def getAreaSet(areaSetBaseName, debugLogPath = ""):
	areaSetName = "aset_%s" % (areaSetBaseName)
	
	aset = None
	if areaSetName in bpy.data.collections.keys():
		WriteToFile(debugLogPath, 'Using existing area set %s' % (areaSetName))
		aset = bpy.data.collections[areaSetName]
	else:
		WriteToFile(debugLogPath, 'Creating area set %s' % (areaSetName))
		aset = bpy.data.collections.new(name=areaSetName)
		bpy.context.scene.collection.children.link(aset)
	return aset

def getAreaWithObjectsContainer(areaName, debugLogPath = ""):
	awo_name = "awo_%s" % (areaName)
	awo = None
	if awo_name in bpy.data.collections.keys():
		WriteToFile(debugLogPath, 'Using existing area-with-objects %s' % (awo_name))
		awo = bpy.data.collections[awo_name]
	else:
		WriteToFile(debugLogPath, 'Creating area-with-objects %s' % (awo_name))
		awo = bpy.data.collections.new(name=awo_name)	
	return awo
	
def getUnplacedObjectAWO():
	awo = getAreaWithObjectsContainer("UnusedObjects")
	bpy.context.scene.collection.children.link(awo)
	return awo

def importArea(areaSetBaseName, areaName, areaColladaPath, placementOrder, distanceFromOrigin, posX, posY, posZ, debugLogPath = ""):
	aset = getAreaSet(areaSetBaseName, debugLogPath)
	awo = getAreaWithObjectsContainer(areaName, debugLogPath)
	
	try:
		preimport_objects = set(bpy.context.scene.objects)
		bpy.ops.wm.collada_import(filepath=areaColladaPath)
		imported_objects = set(bpy.context.scene.objects) - preimport_objects
		if imported_objects and len(imported_objects) > 0:
			current_area_object = getObjectWithShortestName(imported_objects)
			original_area_tlo_name = current_area_object.name
			current_area_object.name = "area_%s" % (areaName)
			renameChildObjects(imported_objects, current_area_object.name, current_area_object.name)
			current_area_object['scripted_placement_order'] = placementOrder
			current_area_object['distance_from_origin'] = distanceFromOrigin
			current_area_object['g2_type'] = 'area'
			current_area_object.location[0] = posX
			current_area_object.location[1] = posY
			current_area_object.location[2] = posZ
			current_area_object.rotation_euler[0] = 0
			current_area_object.rotation_euler[1] = 0
			current_area_object.rotation_euler[2] = 0
			awo.objects.link(bpy.data.objects[current_area_object.name])
	except Exception as e:
		errMessage = 'Error importing "%s": %s' % (areaColladaPath, e)
		WriteToFile(debugLogPath, errMessage)
		print(errMessage)
	try:
		bpy.data.collections[aset.name].children.link(awo)
	except Exception as e:
		errMessage = 'Error linking area-with-objects "%s": %s' % (awo.name, e)
		WriteToFile(debugLogPath, errMessage)
	return awo

#toggle_expand(2)
#setAllMaterialsDefaultAlphaTo1()
#adjustMaterials()
"""
				
	outputFile.write(scriptFunctions)
	outputFile.write(gLineSeparator)
	outputFile.close()
	
	if modelExTestCaseMode:
		meStartX = -100
		meMaxX = 100
		meStartY = 50
		meStepX = 50
		meStepY = -50
		meCurrentX = meStartX
		meCurrentY = meStartY
		itemNum = 1
		testcases = [ "city2", "cliff1", "intvaly1", "nightb3", "tower1", "tower8", "train9", "raziel", "walboss" ]
		WriteToFile(outputFilePath, "# ModelEx test cases")
		for tc in testcases:
			WriteToFile(outputFilePath, "awo = importArea('%s', '%s', '%s', %s, %s, %s, %s, %s, debugLogPath = '%s')" % (tc, tc, os.path.join(areasColladaDirectoryPath, "..", "Export_Test", "%s.dae" % (tc)).replace("\\", "\\\\"), itemNum, itemNum, meCurrentX, meCurrentY, 0.0, logPath.replace("\\", "\\\\")))
			itemNum += 1
			meCurrentX += meStepX
			if meCurrentX > meMaxX:
				meCurrentX = meStartX
				meCurrentY += meStepY
			# "%MODELEX%" --input "%BASEPATH%\%BUILDDATE%\Extracted\kain2\area\cliff\bin\cliff1.pcm" --output "%BASEPATH%\%BUILDDATE%\Export_Test\cliff1.dae" --discard-non-visible --ignore-pflag2-for-terrain
		# awo = importArea("add", "add1", "X:\\Legacy_of_Kain\\Soul_Reaver_Alpha\\1999-06-10\\AreaModelsMaterial\\add1.dae", 204, 31, -53.453, -103.08500000000001, -0.529, debugLogPath = "X:\\Legacy_of_Kain\\Soul_Reaver_Alpha\\1999-06-10\\Blender_Import_Debug_Log-Material-No_Objects.txt")
	else:
	
		unplacedAreas = []
		usedAreaColladaFiles = []
		currentupaX = unlinkedAreaMinX
		currentupaY = unlinkedAreaStartY 
		currentupaZ = unlinkedAreaStartZ
		
		if placeAreas:	
			newArea = AreaFromFile(startArea, inputFilePath, areasColladaDirectoryPath, objectsColladaDirectoryPath, pTracker.AllObjectColladaFiles, pTracker.ImportObjects, pTracker.IgnoredObjectNames, pTracker.ObjectNameWhitelist, pTracker.Counter, pTracker.RandomizedObjectNames, placeObjectsSimultaneously = pTracker.PlaceObjectsSimultaneously)
			newArea.Name = startArea
			newArea.DistanceFromOrigin = initialDistanceFromOrigin
			if startAreaX:
				newArea.PosX = startAreaX
			if startAreaY:
				newArea.PosY = startAreaY
			if startAreaZ:
				newArea.PosZ = startAreaZ
			
			pTracker.RecursePortals(newArea, initialDistanceFromOrigin)
			HandleNewArea(pTracker)
			
		
			# if the option to add unlinked areas is enabled, add each of those one by one and see
			# if they bring anything else along with them
			if placeUnlinkedAreas:
				for areaName in pTracker.Areas.keys():
					checkArea = pTracker.Areas[areaName]
					usedAreaColladaFiles.append(checkArea.ColladaFilePath)
				for colladaPath in pTracker.AllAreaColladaFiles:
					if colladaPath not in usedAreaColladaFiles:
						unplacedAreaName = ntpath.basename(colladaPath)
						if unplacedAreaName.endswith(".dae"):
							unplacedAreaName = unplacedAreaName[:-4]
						unplacedAreas.append(unplacedAreaName)
				unplacedAreas = SortStringsIntuitively(unplacedAreas)
				donePlacing = False		
				if len(unplacedAreas) > 0:
					while not donePlacing:
						nextUnplacedAreaName = unplacedAreas.pop(0)
						linkUnlinkedArea = True
						if nextUnplacedAreaName in pTracker.Areas.keys():
							print("Skipping unlinked area '%s', because it was already placed via linking to another unlinked area" % (nextUnplacedAreaName))
							linkUnlinkedArea = False
						if pTracker.AreaWhitelistRegex is not None and not re.match(pTracker.AreaWhitelistRegex, nextUnplacedAreaName):
							print("Ignoring unlinked area '%s' because it does not match the whitelist regex" % (nextUnplacedAreaName))
							linkUnlinkedArea = False
						if linkUnlinkedArea:
							unlinkedAreaPath = os.path.join(pTracker.AreaJSONDirectory, "%s.json" % nextUnplacedAreaName)
							additionalArea = AreaFromFile(nextUnplacedAreaName, unlinkedAreaPath, areasColladaDirectoryPath, objectsColladaDirectoryPath, pTracker.AllObjectColladaFiles, pTracker.ImportObjects, pTracker.IgnoredObjectNames, pTracker.ObjectNameWhitelist, pTracker.Counter, pTracker.RandomizedObjectNames, placeObjectsSimultaneously = pTracker.PlaceObjectsSimultaneously)
							additionalArea.Name = nextUnplacedAreaName
							additionalArea.DistanceFromOrigin = pTracker.MaxDistanceFromOrigin + 1
							additionalArea.PosX = currentupaX
							additionalArea.PosY = currentupaY
							additionalArea.PosZ = currentupaZ
							print("Placing unlinked area '%s' from '%s' at %s, %s, %s" % (nextUnplacedAreaName, unlinkedAreaPath, additionalArea.PosX, additionalArea.PosY, additionalArea.PosZ))
							pTracker.RecursePortals(additionalArea, additionalArea.DistanceFromOrigin)
							HandleNewArea(pTracker)
							currentupaX += unlinkedAreaStepX
							if currentupaX > unlinkedAreaMaxX:
								currentupaX = unlinkedAreaMinX 
								currentupaY += unlinkedAreaStepY
								currentupaZ += unlinkedAreaStepZ 
						if len(unplacedAreas) == 0:
							donePlacing = True
		
			# Create the collections
			numAreas = len(pTracker.AreaNames)
			# "warpgate" is a quasi-area that doesn't have a DRM, and it messes up the total area count
			if "warpgate" in pTracker.AreaNames:
				numAreas -= 1
			areaNum = 1
			# Loop through all of the areas and add them to a hashtable so that the list of keys can be sorted and then 
			# the actual Blender code created in that order, so that in the Blender tree view, the areas are sorted 
			# properly instead of by the order they were added in.
			# This may seem like a little thing, but it's really frustrating to find individual meshes in the tree view 
			# when the areas are not sorted alphabetically.
			areaBlenderCode = {}
			areaBlenderCodeKeys = []
			warnedAreaSets = []
			warnedAreas = []
			for areaSetName in pTracker.AreaSetNames:
				print('Building area set "%s"' % (areaSetName))
				for areaName in pTracker.AreaNames:
					if areaName in pTracker.Areas.keys():
						area = pTracker.Areas[areaName]
						if area.AreaSetName not in pTracker.AreaSetNames and area.AreaSetName not in warnedAreaSets:
							print("Error: did not find area set name '%s' (for area '%s') in the list of area sets." % (area.AreaSetName, areaName))
							warnedAreaSets.append(area.AreaSetName)
						if area.AreaSetName == areaSetName:
							print('\tAdding area "%s" to area set "%s"' % (areaName, areaSetName))
							#WriteToFile(outputFilePath, "bpy.context.scene.frame_set(%s)" % (frameNumber))
							#frameNumber += frameStep
							
							areaBlenderCode[areaName] = area.GetBlenderCode(positionMultiplier, pTracker.SpectralMode)
							areaBlenderCodeKeys.append(areaName)
							#WriteToFile(outputFilePath, area.GetBlenderCode(positionMultiplier, pTracker.SpectralMode))
							
							if len(area.G2EngineObjects) > 0:
								print("\t\tAdded import code for %s object(s)" % (len(area.G2EngineObjects)))
					else:
						if areaName not in warnedAreas:
							if areaName == "warpgate":
								print("Did not find area '%s' in the list of keys. This is expected." % (areaName))
							else:
								print("Error: did not find area '%s' in the list of keys." % (areaName))
							warnedAreas.append(areaName)

			areaBlenderCodeKeys = SortStringsIntuitively(areaBlenderCodeKeys)
			for areaKey in areaBlenderCodeKeys:
				WriteToFile(outputFilePath, "print('Importing area %s (%s/%s)')" % (areaKey, areaNum, numAreas))
				WriteToFile(outputFilePath, areaBlenderCode[areaKey])
				areaNum += 1
		
		unplacedObjects = []
		usedObjectColladaFiles = []
		for areaName in pTracker.Areas.keys():
			checkArea = pTracker.Areas[areaName]
			for o in checkArea.G2EngineObjects:
				usedObjectColladaFiles.append(o.ColladaFilePath.replace("\\\\", "\\"))
		for colladaPath in pTracker.AllObjectColladaFiles:
			if colladaPath not in usedObjectColladaFiles and colladaPath not in unplacedObjects:
				unplacedObjects.append(colladaPath)

		numUnplacedObjects = len(unplacedObjects)
		if placeUnusedObjects and numUnplacedObjects > 0:
			unplacedObjects = SortStringsIntuitively(unplacedObjects)
			WriteToFile(outputFilePath, "print('Importing unused objects')")
			# reset to next unplaced area-sized row
			WriteToFile(outputFilePath, "awo_unused = getUnplacedObjectAWO()")
			currentupaX = unlinkedAreaMinX 
			if len(pTracker.AreaNames) > 0:
				currentupaY += unlinkedAreaStepY
				currentupaZ += unlinkedAreaStepZ 
			donePlacing = False
			uoNum = 1
			revisedUnplacedObjects = []
			for uo in unplacedObjects:
				newObject = G2EngineObject()
				newObject.ObjectTypeName = ntpath.basename(uo).split(".")[0].replace("-", "_")
				newObject.ColladaFilePath = uo
				newObject.ParentAreaName = "UnusedObjects"
				newObject.DistanceFromOrigin = pTracker.MaxDistanceFromOrigin + 10
				newObject.ParentAreaBlenderName = "UnusedObjects"
				newObject.ObjectNumWithinParent = uoNum
				newObject.Name = "%s_obj_%s_%s" % (newObject.ParentAreaName, newObject.ObjectNumWithinParent, newObject.ObjectTypeName)			
				objectBaseName = newObject.ObjectTypeName.split("_")[0]
				if objectBaseName in pTracker.IgnoredObjectNames and 1 == 2:
					print("Skipping unused object '%s' from '%s', because its type is in the list to ignore" % (newObject.ObjectTypeName, uo))
					revisedUnplacedObjects.append(uo)
				else:
					print("Placing unused object '%s' from '%s' at %s, %s, %s" % (newObject.ObjectTypeName, uo, currentupaX, currentupaY, currentupaZ))
					#WriteToFile(outputFilePath, "print(\"Importing unused object '%s' from '%s'\")" % (newObject.ObjectTypeName, uo.replace("\\", "\\\\")))
					WriteToFile(outputFilePath, newObject.GetBlenderCode(currentupaX, currentupaY, currentupaZ, positionMultiplier, pTracker.SpectralMode, "awo_unused") )
					usedObjectColladaFiles.append(uo)
					
					currentupaX += unusedObjectStepX
					if currentupaX > unlinkedAreaMaxX:
						currentupaX = unlinkedAreaMinX 
						currentupaY += unusedObjectStepY
						currentupaZ += unusedObjectStepZ 
				uoNum += 1
			unplacedObjects = revisedUnplacedObjects
		
		if len(usedObjectColladaFiles) > 0:
			usedObjectColladaFiles = SortStringsIntuitively(usedObjectColladaFiles)
			objectCounts = {}
			for o in usedObjectColladaFiles:
				if o in objectCounts.keys():
					objectCounts[o] += 1
				else:
					objectCounts[o] = 1
			print("Placed the following object files:")
			listedFiles = []
			for o in usedObjectColladaFiles:
				#print(o)
				if o not in listedFiles:
					print("'%s': %s time(s)" % (o, objectCounts[o]))
					listedFiles.append(o)
		if len(unplacedObjects) > 0:
			print("Warning: the following object files were found in the model directory, but not placed by the automatic linking process:")
			unplacedObjects = SortStringsIntuitively(unplacedObjects)
			for uo in unplacedObjects:
				print(uo)

		unplacedAreas = []
		usedAreaColladaFiles = []
		for areaName in pTracker.Areas.keys():
			checkArea = pTracker.Areas[areaName]
			usedAreaColladaFiles.append(checkArea.ColladaFilePath)
		for colladaPath in pTracker.AllAreaColladaFiles:
			if colladaPath not in usedAreaColladaFiles:
				unplacedAreas.append(colladaPath)
		
		if len(usedAreaColladaFiles) > 0:
			usedAreaColladaFiles = SortStringsIntuitively(usedAreaColladaFiles)
			print("Placed the following area files:")
			for a in usedAreaColladaFiles:
				print(a)
		if len(unplacedAreas) > 0:
			print("Warning: the following area files were found in the model directory, but not placed by the automatic linking process:")
			unplacedAreas = SortStringsIntuitively(unplacedAreas)
			for ua in unplacedAreas:
				print(ua)

		print("Max distance from origin: %s" % (pTracker.MaxDistanceFromOrigin))

	#post-import steps
	WriteToFile(outputFilePath, "## Delete empty objects")
	WriteToFile(outputFilePath, "#deleteEmptyObjects(debugLogPath = '%s')" % (logPath.replace("\\", "\\\\")))
	WriteToFile(outputFilePath, "# Create missing vertex colour layers")
	WriteToFile(outputFilePath, "createMissingVertexColours(debugLogPath = '%s')" % (logPath.replace("\\", "\\\\")))
	WriteToFile(outputFilePath, "## Delete orphaned materials")
	WriteToFile(outputFilePath, "#deleteOrphanedMaterials(debugLogPath = '%s')" % (logPath.replace("\\", "\\\\")))
	WriteToFile(outputFilePath, "# collapse all collection entries in all contexts/views for convenience")
	WriteToFile(outputFilePath, "toggle_expand(2)")
	WriteToFile(outputFilePath, "# adjust materials for all objects")
	WriteToFile(outputFilePath, "adjustMaterials(debugLogPath = '%s', emulateBackfaceCulling = %s, everythingGlows = %s)" % (logPath.replace("\\", "\\\\"), emulateBackfaceCulling, everythingGlows))
	
	WriteToFile(outputFilePath, "#Set ambient lighting")
	if pTracker.SpectralLightingMode:
		WriteToFile(outputFilePath, 'bpy.data.worlds["World"].node_tree.nodes["Background"].inputs[0].default_value = (0.0156862, 0.172549, 0.172549, 1.0)')
	else:
		#WriteToFile(outputFilePath, 'bpy.data.worlds["World"].node_tree.nodes["Background"].inputs[0].default_value = (0.226584, 0.226584, 0.226584, 1.0)')
		WriteToFile(outputFilePath, 'bpy.data.worlds["World"].node_tree.nodes["Background"].inputs[0].default_value = (0.050876, 0.050876, 0.050876, 1.0)')
	if addSpinCam or addTripleTierSpinCam:
		WriteToFile(outputFilePath, """# Add spincam
cam1 = getNewCamera("spincam")
targetCube = getNewCube()
targetCube.scale[0] = 0.0001
targetCube.scale[1] = 0.0001
targetCube.scale[2] = 0.0001
""")
	if addSpinCam:
		expString = ""
		if exponentialSpin:
			expString = "useExponentialSpeedForSpin = True, useExponentialSpeedForDistance = True, useExponentialSpeedForHeight = True,"
		WriteToFile(outputFilePath, "doSpinCam(cam1, targetCube, %s, %s, %s, %s, %s, %s, %s, %s speedExponent = 1.5, debugLogPath = '%s')" % (spinCamNumRotations, spinCamStartDistance, spinCamEndDistance, spinCamStartHeight, spinCamEndHeight, 0, spinCamRotationFrameCount - 1, expString, logPath.replace("\\", "\\\\")))
	
	if addTripleTierSpinCam:
		WriteToFile(outputFilePath, "doTripleTierSpinCam(cam1, targetCube, %s, %s, %s, %s, %s, %s, %s, %s, debugLogPath = '%s')" % (ttSpinCamTier1Distance, ttSpinCamTier2Distance, ttSpinCamTier3Distance, ttSpinCamTier1Height, ttSpinCamTier2Height, ttSpinCamTier3Height, 0, spinCamRotationFrameCount - 1, logPath.replace("\\", "\\\\")))

	if addSpinCam or addTripleTierSpinCam:
		WriteToFile(outputFilePath, "# Reset to initial frame")
		WriteToFile(outputFilePath, "bpy.context.scene.frame_set(0)")

	if addOverheadLights:
		cYScaled = float(currentupaY) * positionMultiplier
		if cYScaled < overheadLightMinY:
			overheadLightMinY = cYScaled - (cYScaled % overheadLightStepY) - (overheadLightStepY * 4.0)
		WriteToFile(outputFilePath, "# Add overhead light array")
		lightColour = "(1.0, 1.0, 1.0)"
		if pTracker.SpectralLightingMode:
			lightColour = "(0.0156862, 0.172549, 0.172549)"
		WriteToFile(outputFilePath, "addOverheadLightArray(%s, %s, %s, %s, %s, %s, %s, %s, %s, debugLogPath = '%s')" % (overheadLightMinX, overheadLightMinY, overheadLightMaxX, overheadLightMaxY, overheadLightStepX, overheadLightStepY, overheadLightZ, overheadLightStrength, lightColour, logPath.replace("\\", "\\\\")))
#		WriteToFile(outputFilePath, """# Add sun
#newLight = getNewLight("overhead_sun", 'SUN', lightStrength = 10.0)
#newLight.location[0] = 0.0
#newLight.location[1] = 0.0
#newLight.location[2] = 412.23
#""")
#		if pTracker.SpectralMode:
#			WriteToFile(outputFilePath, "newLight.data.color = (0.226584, 0.226584, 0.226584)")
	
	WriteToFile(outputFilePath, "# Arrange all shader nodes for readability")
	WriteToFile(outputFilePath, "arrangeAllShaderNodes(debugLogPath = '%s')" % (logPath.replace("\\", "\\\\")))
	
	WriteToFile(outputFilePath, "# collapse all collection entries in all contexts/views for convenience")
	WriteToFile(outputFilePath, "toggle_expand(2)")
	#WriteToFile(outputFilePath, "# adjust materials for all objects")
	#WriteToFile(outputFilePath, "adjustMaterials(debugLogPath = '%s')" % (logPath.replace("\\", "\\\\")))
	#WriteToFile(outputFilePath, "toggle_expand(2)")
	
	# Add overhead orthographic camera(s)
	WriteToFile(outputFilePath, """# Set ambient light to 25% intensity - use this for overhead ortho, otherwise 1.0 is correct.
#bpy.data.worlds["World"].node_tree.nodes["Background"].inputs[1].default_value = 0.25
	
# Add overhead orthographic cameras
ocam1 = getNewCamera("overhead_ortho1")
ocam1.data.type = 'ORTHO'
ocam1.location = (0, 0, 1000.0)
ocam1.rotation_euler = (0, 0, 0)
ocam1.data.clip_end = 100000
ocam1.data.ortho_scale = 650.0

ocam2 = getNewCamera("overhead_ortho2")
ocam2.data.type = 'ORTHO'
ocam2.location = (0, -270.0, 1000.0)
ocam2.rotation_euler = (0, 0, 0)
ocam2.data.clip_end = 100000
ocam2.data.ortho_scale = 650.0

ocam3 = getNewCamera("overhead_ortho3")
ocam3.data.type = 'ORTHO'
ocam3.location = (0, -540.0, 1000.0)
ocam3.rotation_euler = (0, 0, 0)
ocam3.data.clip_end = 100000
ocam3.data.ortho_scale = 650.0

# Add general-purpose cam that's in a reasonably-convenient location for making assembly animations
acam1 = getNewCamera("assembly_cam")
acam1.location = (0.0, -175.0, 130.0)
acam1.rotation_euler = (0.93043, 0.0, 0.0)
acam1.data.clip_end = 1000""")
	# set Cycles as the renderer
	WriteToFile(outputFilePath, "# set Cycles as the renderer, because the number of materials is too high for Eevee")
	WriteToFile(outputFilePath, "bpy.context.scene.render.engine = '%s'" % (renderEngine))
	
	# Set start frame to 0
	WriteToFile(outputFilePath, """# Set start frame to 0
bpy.context.scene.frame_start = 0
bpy.context.scene.frame_set(0)""")
		
	WriteToFile(outputFilePath, "print('Done importing')")
		
